__author__ = "Simon Nilsson"

import os
import shutil
import subprocess
from datetime import datetime
from typing import Dict, Optional

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

from simba.mixins.config_reader import ConfigReader
from simba.utils.checks import check_nvidea_gpu_available
from simba.utils.enums import Paths, TagNames
from simba.utils.errors import FFMPEGCodecGPUError
from simba.utils.printing import SimbaTimer, log_event, stdout_success
from simba.utils.read_write import (get_fn_ext, get_video_meta_data,
                                    read_config_entry, read_config_file,
                                    remove_a_folder)


class FrameMergererFFmpeg(ConfigReader):
    """
    Merge separate visualizations of classifications, descriptive statistics etc., into  single video mosaic.

    .. note::
       `GitHub tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/tutorial.md#step-12-merge-frames>`_.

        .. image:: _static/img/mosaic_videos.gif
          :width: 600
          :align: center

    :parameter str config_path: path to SimBA project config file in Configparser format
    :parameter str concat_type: Type of concatenation. OPTIONS: 'horizontal', 'vertical', 'mosaic', 'mixed_mosaic'.
    :parameter dict frame_types: Dict holding video path to videos to concatenate. E.g., {'Video 1': path, 'Video 2': path}
    :parameter int video_height: Output video height.
    :parameter int video_width: Output video width.
    :parameter Optional[bool] gpu: If True, use NVIDEA FFMpeg GPU codecs. Default False.

    Example
    ----------
    >>> frame_types={'Video 1': 'project_folder/videos/Video_1.avi', 'Video 2': 'project_folder/videos/Video_2.avi'}
    >>> video_height, video_width, concat_type = 640, 480, 'vertical'
    >>> FrameMergererFFmpeg(config_path='MySimBaConfigPath', frame_types=frame_types, video_height=video_height, video_width=video_width, concat_type=concat_type)
    """

    def __init__(
        self,
        concat_type: Literal["horizontal", "vertical", "mosaic", "mixed_mosaic"],
        frame_types: Dict[str, str],
        video_height: int,
        video_width: int,
        config_path: Optional[str] = None,
        gpu: Optional[bool] = False,
    ):
        if gpu and not check_nvidea_gpu_available():
            raise FFMPEGCodecGPUError(
                msg="NVIDEA GPU not available (as evaluated by nvidea-smi returning None",
                source=self.__class__.__name__,
            )
        self.timer = SimbaTimer(start=True)
        self.datetime = datetime.now().strftime("%Y%m%d%H%M%S")
        if config_path is not None:
            ConfigReader.__init__(self, config_path=config_path)
            log_event(
                logger_name=str(__class__.__name__),
                log_type=TagNames.CLASS_INIT.value,
                msg=self.create_log_msg_from_init_args(locals=locals()),
            )
            self.output_dir = os.path.join(
                self.project_path, Paths.CONCAT_VIDEOS_DIR.value
            )
            self.temp_dir = os.path.join(
                self.project_path, Paths.CONCAT_VIDEOS_DIR.value, "temp"
            )
            self.output_path = os.path.join(
                self.project_path,
                Paths.CONCAT_VIDEOS_DIR.value,
                f"merged_video_{self.datetime}.mp4",
            )
        else:
            self.file_path = list(frame_types.values())[0]
            self.output_dir, ss, df = get_fn_ext(filepath=self.file_path)
            self.temp_dir = os.path.join(self.output_dir, "temp")
            self.output_path = os.path.join(
                self.output_dir, f"merged_video_{self.datetime}.mp4"
            )

        self.video_height, self.video_width, self.gpu = video_height, video_width, gpu
        self.frame_types, self.concat_type = frame_types, concat_type
        self.video_cnt = len(self.frame_types.keys())
        self.even_bool = (len(self.frame_types.keys()) % 2) == 0
        self.blank_path = os.path.join(self.temp_dir, "blank.mp4")
        if not os.path.exists(self.output_dir):
            os.makedirs(self.output_dir)
        if not os.path.exists(self.temp_dir):
            os.makedirs(self.temp_dir)
        if concat_type == "horizontal":
            self.__horizontal_concatenator(
                out_path=self.output_path, frames_dict=self.frame_types, final_img=True
            )
        elif concat_type == "vertical":
            self.__vertical_concatenator(
                out_path=self.output_path, frames_dict=self.frame_types, final_img=True
            )
        elif concat_type == "mosaic":
            self.__mosaic_concatenator(
                output_path=self.output_path,
                frames_dict=self.frame_types,
                final_img=True,
            )
        elif concat_type == "mixed_mosaic":
            self.__mixed_mosaic_concatenator()
        remove_a_folder(folder_dir=self.temp_dir)

    def __resize_height(self, new_height: int):
        for video_cnt, (video_type, video_path) in enumerate(self.frame_types.items()):
            video_meta_data = get_video_meta_data(video_path=video_path)
            out_path = os.path.join(self.temp_dir, video_type + ".mp4")
            if video_meta_data["height"] != new_height:
                print("Resizing {}...".format(video_type))
                if self.gpu:
                    command = 'ffmpeg -y -hwaccel auto -c:v h264_cuvid -i "{}" -vf scale_npp=-2:{} -c:v h264_nvenc "{}" -hide_banner -loglevel error -y'.format(
                        video_path, new_height, out_path
                    )
                else:
                    command = 'ffmpeg -y -i "{}" -vf scale=-2:{} "{}" -hide_banner -loglevel error -y'.format(
                        video_path, new_height, out_path
                    )
                subprocess.call(command, shell=True, stdout=subprocess.PIPE)
            else:
                shutil.copy(video_path, out_path)

    def __resize_width(self, new_width: int):
        """Helper to change the width of videos"""

        for video_cnt, (video_type, video_path) in enumerate(self.frame_types.items()):
            video_meta_data = get_video_meta_data(video_path=video_path)
            out_path = os.path.join(self.temp_dir, video_type + ".mp4")
            if video_meta_data["height"] != new_width:
                print("Resizing {}...".format(video_type))
                if self.gpu:
                    command = 'ffmpeg -y -hwaccel auto -i "{}" -vf scale_npp={}:-2 -c:v h264_nvenc "{}" -hide_banner -loglevel error -y'.format(
                        video_path, new_width, out_path
                    )
                else:
                    command = 'ffmpeg -y -i "{}" -vf scale={}:-2 "{}" -hide_banner -loglevel error -y'.format(
                        video_path, new_width, out_path
                    )
                subprocess.call(command, shell=True, stdout=subprocess.PIPE)
            else:
                shutil.copy(video_path, out_path)

    def __resize_width_and_height(self, new_width: int, new_height: int):
        """Helper to change the width and height videos"""

        for video_cnt, (video_type, video_path) in enumerate(self.frame_types.items()):
            video_meta_data = get_video_meta_data(video_path=video_path)
            out_path = os.path.join(self.temp_dir, video_type + ".mp4")
            if video_meta_data["height"] != new_width:
                print("Resizing {}...".format(video_type))
                if self.gpu:
                    command = f'ffmpeg -y -hwaccel auto -c:v h264_cuvid -i "{video_path}" -vf "scale=w={new_width}:h={new_height}:force_original_aspect_ratio=decrease:flags=bicubic" -c:v h264_nvenc "{out_path}" -hide_banner -loglevel error -y'
                else:
                    command = 'ffmpeg -y -i "{}" -vf scale={}:{} "{}" -hide_banner -loglevel error -y'.format(
                        video_path, new_width, new_height, out_path
                    )
                subprocess.call(command, shell=True, stdout=subprocess.PIPE)
            else:
                shutil.copy(video_path, out_path)

    def __create_blank_video(self):
        """Helper to create a blank (black) video"""
        video_meta_data = get_video_meta_data(list(self.frame_types.values())[0])
        if self.gpu:
            cmd = 'ffmpeg -y -t {} -f lavfi -i color=c=black:s={}x{} -c:v h264_nvenc -preset slow -tune stillimage -pix_fmt yuv420p "{}" -hide_banner -loglevel error -y'.format(
                str(video_meta_data["video_length_s"]),
                str(self.video_width),
                str(self.video_height),
                self.blank_path,
            )
        else:
            cmd = 'ffmpeg -y -t {} -f lavfi -i color=c=black:s={}x{} -c:v libx264 -tune stillimage -pix_fmt yuv420p "{}" -hide_banner -loglevel error -y'.format(
                str(video_meta_data["video_length_s"]),
                str(self.video_width),
                str(self.video_height),
                self.blank_path,
            )
        subprocess.call(cmd, shell=True, stdout=subprocess.PIPE)
        self.frame_types["blank"] = self.blank_path

    def __horizontal_concatenator(
        self, frames_dict: dict, out_path: str, include_resize=True, final_img=True
    ):
        """Helper to horizontally concatenate N videos"""

        if include_resize:
            self.__resize_height(new_height=self.video_height)
        video_path_str = ""
        for video_type in frames_dict.keys():
            video_path_str += ' -i "{}"'.format(
                os.path.join(self.temp_dir, video_type + ".mp4")
            )
        if self.gpu:
            cmd = 'ffmpeg -y{} -filter_complex "hstack=inputs={}" -c:v h264_nvenc -vsync 2 "{}" -hide_banner -loglevel error -y'.format(
                video_path_str, str(len(frames_dict.keys())), out_path
            )
        else:
            cmd = 'ffmpeg -y{} -filter_complex hstack=inputs={} -vsync 2 "{}" -hide_banner -loglevel error -y'.format(
                video_path_str, str(len(frames_dict.keys())), out_path
            )
        print(
            "Concatenating (horizontal) {} videos...".format(
                str(len(frames_dict.keys()))
            )
        )
        subprocess.call(cmd, shell=True, stdout=subprocess.PIPE)
        if final_img:
            self.timer.stop_timer()
            stdout_success(
                msg=f"Merged video saved at {out_path}",
                elapsed_time=self.timer.elapsed_time_str,
                source=self.__class__.__name__,
            )

    def __vertical_concatenator(
        self, frames_dict: dict, out_path: str, include_resize=True, final_img=True
    ):
        """Helper to vertically concatenate N videos"""

        if include_resize:
            self.__resize_width(new_width=self.video_width)
        video_path_str = ""
        for video_type in frames_dict.keys():
            video_path_str += ' -i "{}"'.format(
                os.path.join(self.temp_dir, video_type + ".mp4")
            )
        if self.gpu:
            cmd = 'ffmpeg -y{} -filter_complex "vstack=inputs={}" -c:v h264_nvenc -vsync 2 "{}" -hide_banner -loglevel error -y'.format(
                video_path_str, str(len(frames_dict.keys())), out_path
            )
        else:
            cmd = 'ffmpeg -y{} -filter_complex vstack=inputs={} -vsync 2 "{}" -hide_banner -loglevel error -y'.format(
                video_path_str, str(len(frames_dict.keys())), out_path
            )
        print(
            "Concatenating (vertical) {} videos...".format(str(len(frames_dict.keys())))
        )
        subprocess.call(cmd, shell=True, stdout=subprocess.PIPE)
        if final_img:
            self.timer.stop_timer()
            stdout_success(
                msg=f"Merged video saved at {out_path}",
                elapsed_time=self.timer.elapsed_time_str,
                source=self.__class__.__name__,
            )

    def __mosaic_concatenator(
        self, frames_dict: dict, output_path: str, final_img: bool
    ):
        self.__resize_width_and_height(
            new_width=self.video_width, new_height=self.video_height
        )
        if not self.even_bool:
            self.__create_blank_video()
        lower_dict = dict(list(frames_dict.items())[len(frames_dict) // 2 :])
        upper_dict = dict(list(frames_dict.items())[: len(frames_dict) // 2])
        if (len(upper_dict.keys()) == 1) & (len(lower_dict.keys()) != 1):
            upper_dict["blank"] = self.blank_path
        if (len(lower_dict.keys()) == 1) & (len(upper_dict.keys()) != 1):
            lower_dict["blank"] = self.blank_path
        self.__horizontal_concatenator(
            frames_dict=upper_dict,
            out_path=os.path.join(self.temp_dir, "upper.mp4"),
            include_resize=False,
            final_img=False,
        )
        self.__horizontal_concatenator(
            frames_dict=lower_dict,
            out_path=os.path.join(self.temp_dir, "lower.mp4"),
            include_resize=False,
            final_img=False,
        )
        frames_dict = {
            "upper": os.path.join(self.temp_dir, "upper.mp4"),
            "lower": os.path.join(self.temp_dir, "lower.mp4"),
        }
        self.__vertical_concatenator(
            frames_dict=frames_dict,
            out_path=output_path,
            include_resize=False,
            final_img=final_img,
        )

    def __mixed_mosaic_concatenator(self):
        large_mosaic_dict = {
            str(list(self.frame_types.keys())[0]): str(
                list(self.frame_types.values())[0]
            )
        }
        del self.frame_types[str(list(self.frame_types.keys())[0])]
        self.even_bool = (len(self.frame_types.keys()) % 2) == 0
        output_path = os.path.join(self.temp_dir, "mosaic.mp4")
        self.__mosaic_concatenator(
            frames_dict=self.frame_types, output_path=output_path, final_img=False
        )
        self.frame_types = large_mosaic_dict
        self.__resize_height(
            new_height=get_video_meta_data(video_path=output_path)["height"]
        )
        self.frame_types = {
            str(list(large_mosaic_dict.keys())[0]): os.path.join(
                self.temp_dir, str(list(large_mosaic_dict.keys())[0]) + ".mp4"
            ),
            "mosaic": os.path.join(self.temp_dir, "mosaic.mp4"),
        }
        self.__horizontal_concatenator(
            frames_dict=self.frame_types,
            out_path=self.output_path,
            include_resize=False,
            final_img=True,
        )


# FrameMergererFFmpeg(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                     frame_types={'Video 1': '/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/videos/Together_1.avi',
#                                  'Video 2': '/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/videos/Together_1.avi'},
#                     video_height=640,
#                     video_width=480,
#                     concat_type='vertical') #horizontal, vertical, mosaic, mixed_mosaic
#
#
# FrameMergererFFmpeg(config_path=None,
#                     frame_types={'Video 1': '/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/videos/Together_1.avi',
#                                  'Video 2': '/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/videos/Together_1.avi'},
#                     video_height=640,
#                     video_width=480,
#                     concat_type='vertical') #horizontal, vertical, mosaic, mixed_mosaic
#
#


# FrameMergererFFmpeg(config_path=None,
#                     frame_types={'Video 1': r'C:\Users\Nape_Computer_2\Desktop\test_videos\Box1_PM2_day_5_20211104T171021.mp4',
#                                  'Video 2': r'C:\Users\Nape_Computer_2\Desktop\test_videos\Box1_PM2_day_5_20211104T171021.mp4',
#                                  'Video 3': r'C:\Users\Nape_Computer_2\Desktop\test_videos\Box1_PM2_day_5_20211104T171021.mp4',
#                                  'Video 4': r'C:\Users\Nape_Computer_2\Desktop\test_videos\Box1_PM2_day_5_20211104T171021.mp4'},
#                     video_height=640,
#                     video_width=480,
#                     concat_type='mixed_mosaic',
#                     gpu=False) #horizontal, vertical, mosaic, mixed_mosaic
#



__author__ = "Simon Nilsson"

import os.path
from tkinter import *

from simba.feature_extractors.feature_subsets import FeatureSubsetsCalculator
from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.ui.tkinter_functions import (CreateLabelFrameWithIcon, DropDownMenu,
                                        FolderSelect)
from simba.utils.checks import check_if_dir_exists
from simba.utils.enums import Links
from simba.utils.errors import InvalidInputError


class FeatureSubsetExtractorPopUp(PopUpMixin, ConfigReader):
    def __init__(self, config_path: str):
        PopUpMixin.__init__(self, title="EXTRACT FEATURE SUBSETS", size=(500, 500))
        ConfigReader.__init__(self, config_path=config_path)
        self.feature_subset_options = [
            "Two-point body-part distances (mm)",
            "Within-animal three-point body-part angles (degrees)",
            "Within-animal three-point convex hull perimeters (mm)",
            "Within-animal four-point convex hull perimeters (mm)",
            "Entire animal convex hull perimeters (mm)",
            "Entire animal convex hull area (mm2)",
            "Frame-by-frame body-part movements (mm)",
            "Frame-by-frame distance to ROI centers (mm)",
            "Frame-by-frame body-parts inside ROIs (Boolean)",
        ]

        self.settings_frm = CreateLabelFrameWithIcon(
            parent=self.main_frm,
            header="SETTINGS",
            icon_name="documentation",
            icon_link=Links.FEATURE_SUBSETS.value,
        )
        self.save_dir = FolderSelect(self.settings_frm, "SAVE DIRECTORY:", lblwidth=20)
        self.append_to_features_extracted_var = BooleanVar(value=False)
        self.append_to_targets_inserted_var = BooleanVar(value=False)
        self.include_file_checks = BooleanVar(value=False)
        self.append_to_features_cb = Checkbutton(
            self.settings_frm,
            text="APPEND RESULTS TO FEATURES EXTRACTED FILES",
            variable=self.append_to_features_extracted_var,
        )
        self.append_to_targets_inserted_cb = Checkbutton(
            self.settings_frm,
            text="APPEND RESULTS TO TARGET INSERTED FILES",
            variable=self.append_to_targets_inserted_var,
        )
        self.include_file_checks_cb = Checkbutton(
            self.settings_frm,
            text="INCLUDE INTEGRITY CHECKS BEFORE APPENDING NEW DATA (RECOMMENDED)",
            variable=self.include_file_checks,
        )
        self.settings_frm.grid(row=0, column=0, sticky=NW)
        self.save_dir.grid(row=1, column=0, sticky=NW)
        self.append_to_features_cb.grid(row=2, column=0, sticky=NW)
        self.append_to_targets_inserted_cb.grid(row=3, column=0, sticky=NW)
        self.include_file_checks_cb.grid(row=4, column=0, sticky=NW)
        self.feature_subset_selections = self.create_cb_frame(
            main_frm=self.main_frm,
            cb_titles=self.feature_subset_options,
            frm_title="FEATURE SUB-SETS",
        )
        self.create_run_frm(run_function=self.run)
        # self.main_frm.mainloop()

    def run(self):
        selected_features = []
        for feature_name, feature_var in self.feature_subset_selections.items():
            if feature_var.get():
                selected_features.append(feature_name)
        if len(selected_features) == 0:
            raise InvalidInputError(
                msg="Please select at least ONE feature subset family.",
                source=self.__class__.__name__,
            )
        if (
            not self.append_to_targets_inserted_var.get()
            and not self.append_to_features_extracted_var.get()
        ):
            if not os.path.isdir(self.save_dir.folder_path):
                raise InvalidInputError(
                    msg="You must select a valid path where to save the results OR select to append the features to the targets_inserted and/or features_extracted files.",
                    source=self.__class__.__name__,
                )

        feature_extractor = FeatureSubsetsCalculator(
            config_path=self.config_path,
            feature_families=selected_features,
            save_dir=self.save_dir.folder_path,
            include_file_checks=self.include_file_checks.get(),
            append_to_features_extracted=self.append_to_features_extracted_var.get(),
            append_to_targets_inserted=self.append_to_targets_inserted_var.get(),
        )
        feature_extractor.run()


# FeatureSubsetExtractorPopUp(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini')



__author__ = "Simon Nilsson"

import glob
import os

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns

from simba.mixins.unsupervised_mixin import UnsupervisedMixin
from simba.unsupervised.enums import Clustering, Unsupervised
from simba.utils.checks import (check_if_dir_exists,
                                check_if_filepath_list_is_empty)
from simba.utils.printing import stdout_success


class GridSearchVisualizer(UnsupervisedMixin):
    """
    Visualize grid-searched hyper-parameters in .png format.

    :param model_dir: path to pickle holding unsupervised results in ``data_map.yaml`` format.
    :param save_dir: directory holding one or more unsupervised results in pickle ``data_map.yaml`` format.
    :param settings: User-defined image attributes (e.g., continous and catehorical palettes)

    :example:
    >>> settings = {'CATEGORICAL_PALETTE': 'Pastel1', 'SCATTER_SIZE': 10}
    >>> visualizer = GridSearchVisualizer(model_dir='/Users/simon/Desktop/envs/troubleshooting/unsupervised/cluster_models_042023', save_dir='/Users/simon/Desktop/envs/troubleshooting/unsupervised/images', settings=settings)
    >>> visualizer.continuous_visualizer(continuous_vars=['START_FRAME'])
    >>> visualizer.categorical_visualizer(categoricals=['CLUSTER'])
    """

    def __init__(self, model_dir: str, save_dir: str, settings: dict):
        super().__init__()
        check_if_dir_exists(in_dir=save_dir)
        check_if_dir_exists(in_dir=model_dir)
        self.save_dir, self.settings, self.model_dir = save_dir, settings, model_dir
        self.data_path = glob.glob(model_dir + "/*.pickle")
        check_if_filepath_list_is_empty(
            filepaths=self.data_path,
            error_msg=f"SIMBA ERROR: No pickle files in {model_dir}",
        )

    def __join_data(self, data: object):
        embedding_data = pd.DataFrame(
            data[Unsupervised.DR_MODEL.value][Unsupervised.MODEL.value].embedding_,
            columns=["X", "Y"],
        )
        bouts_data = data[Unsupervised.DATA.value][Unsupervised.BOUTS_FEATURES.value]
        target_data = data[Unsupervised.DATA.value][Unsupervised.BOUTS_TARGETS.value]
        cluster_data = pd.DataFrame(
            data[Clustering.CLUSTER_MODEL.value][Unsupervised.MODEL.value]
            .labels_.reshape(-1, 1)
            .astype(np.int8),
            columns=["CLUSTER"],
        )
        data = pd.concat(
            [embedding_data, bouts_data, target_data, cluster_data], axis=1
        )
        return data.loc[:, ~data.columns.duplicated()].copy()

    def categorical_visualizer(self, categoricals: list):
        data = self.read_pickle(self.model_dir)
        for k, v in data.items():
            self.check_key_exist_in_object(object=v, key=Unsupervised.DR_MODEL.value)
        for k, v in data.items():
            data = self.__join_data(data=v)
            for variable in categoricals:
                save_path = os.path.join(
                    self.save_dir,
                    f"{v[Unsupervised.DR_MODEL.value][Unsupervised.HASHED_NAME.value]}_{variable}.png",
                )
                if os.path.isfile(save_path):
                    continue
                else:
                    sns.scatterplot(
                        data=data,
                        x="X",
                        y="Y",
                        hue=variable,
                        palette=sns.color_palette(
                            self.settings["CATEGORICAL_PALETTE"],
                            len(data[variable].unique()),
                        ),
                    )
                    plt.savefig(save_path)
                    stdout_success(msg=f"Saved {save_path}...")
                    plt.close("all")
        self.timer.stop_timer()
        stdout_success(
            msg="All cluster images created.", elapsed_time=self.timer.elapsed_time_str
        )

    def continuous_visualizer(self, continuous_vars: list):
        data = self.read_pickle(self.model_dir)
        for k, v in data.items():
            self.check_key_exist_in_object(object=v, key=Unsupervised.DR_MODEL.value)
        for k, v in data.items():
            data = self.__join_data(data=v)
            for variable in continuous_vars:
                save_path = os.path.join(
                    self.save_dir,
                    f"{v[Unsupervised.DR_MODEL.value][Unsupervised.HASHED_NAME.value]}_{variable}.png",
                )
                fig, ax = plt.subplots()
                plt.xlabel("X")
                plt.ylabel("Y")
                points = ax.scatter(
                    data["X"],
                    data["Y"],
                    c=data[variable],
                    s=self.settings["SCATTER_SIZE"],
                    cmap=self.settings["CONTINUOUS_PALETTE"],
                )
                cbar = fig.colorbar(points)
                cbar.set_label(variable, loc="center")
                title = v[Unsupervised.DR_MODEL.value][Unsupervised.HASHED_NAME.value]
                plt.title(
                    title,
                    ha="center",
                    fontsize=15,
                    bbox={"facecolor": "orange", "alpha": 0.5, "pad": 0},
                )
                fig.savefig(save_path)
                plt.close("all")
                stdout_success(msg=f"Saved {save_path}...")
        self.timer.stop_timer()
        stdout_success(
            msg="All cluster images created.", elapsed_time=self.timer.elapsed_time_str
        )


# settings = {'PALETTE': 'Pastel1'}
# test = GridSearchVisualizer(model_dir='/Users/simon/Desktop/envs/troubleshooting/unsupervised/cluster_models',
#                             save_dir='/Users/simon/Desktop/envs/troubleshooting/unsupervised/images',
#                             settings=settings)
# test.cluster_visualizer()


# settings = {'CATEGORICAL_PALETTE': 'Pastel1', 'SCATTER_SIZE': 10}
# test = GridSearchVisualizer(model_dir='/Users/simon/Desktop/envs/troubleshooting/unsupervised/cluster_models_042023',
#                             save_dir='/Users/simon/Desktop/envs/troubleshooting/unsupervised/images',
#                             settings=settings)
# #test.continuous_visualizer(continuous_vars=['START_FRAME'])
# test.categorical_visualizer(categoricals=['CLUSTER'])



__author__ = "Simon Nilsson"

import glob
import os
import shutil
from typing import List, Union

import pandas as pd

from simba.data_processors.interpolation_smoothing import Interpolate, Smooth
from simba.utils.checks import (check_file_exist_and_readable,
                                check_if_dir_exists,
                                check_if_filepath_list_is_empty, check_int)
from simba.utils.data import smooth_data_gaussian, smooth_data_savitzky_golay
from simba.utils.enums import Methods
from simba.utils.errors import FileExistError, NoFilesFoundError
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import (get_fn_ext,
                                    get_number_of_header_columns_in_df,
                                    read_config_file,
                                    read_project_path_and_file_type)


def import_dlc_csv(config_path: Union[str, os.PathLike], source: str) -> List[str]:
    """
    Import file or folder of  DLC pose-estimation CSV files to SimBA project.
    Returns list of file paths that has been imported.

    :parameter str config_path: path to SimBA project config file in Configparser format
    :parameter str source: path to file or folder containing DLC pose-estimation CSV files
    :return List[str]: Paths of imported files.

    :example:
    >>> import_dlc_csv(config_path='project_folder/project_config.ini', source='CSV_import/Together_1.csv')
    >>> ['project_folder/csv/input_csv/Together_1.csv']
    """

    config = read_config_file(config_path=config_path)
    project_path, file_type = read_project_path_and_file_type(config=config)
    original_file_name_dir = os.path.join(
        project_path, "csv", "input_csv", "original_filename"
    )
    input_csv_dir = os.path.join(project_path, "csv", "input_csv")
    imported_files = glob.glob(input_csv_dir + "/*." + file_type)
    imported_file_names, imported_file_paths = [], []
    for file_path in imported_files:
        _, video_name, _ = get_fn_ext(filepath=file_path)
        imported_file_names.append(video_name)
    if not os.path.exists(original_file_name_dir):
        os.makedirs(original_file_name_dir)
    if os.path.isdir(source):
        csv_files = glob.glob(source + "/*.csv")
        check_if_filepath_list_is_empty(
            csv_files,
            error_msg=f"SIMBA ERROR: NO .csv files found in {source} directory.",
        )
    else:
        csv_files = [source]

    for file_path in csv_files:
        video_timer = SimbaTimer(start=True)
        check_file_exist_and_readable(file_path=file_path)
        _, video_name, file_ext = get_fn_ext(filepath=file_path)
        if "DLC_" in video_name:
            new_file_name = video_name.split("DLC_")[0] + ".csv"
        elif "DeepCut" in video_name:
            new_file_name = video_name.split("DeepCut")[0] + ".csv"
        else:
            new_file_name = video_name.split(".")[0] + ".csv"
        new_file_name_wo_ext = new_file_name.split(".")[0]
        video_basename = os.path.basename(file_path)
        print(f"Importing {video_name} to SimBA project...")
        if new_file_name_wo_ext in imported_file_names:
            raise FileExistError(
                "SIMBA IMPORT ERROR: {} already exist in project. Remove file from project or rename imported video file name before importing.".format(
                    new_file_name
                )
            )
        shutil.copy(file_path, input_csv_dir)
        shutil.copy(file_path, original_file_name_dir)
        os.rename(
            os.path.join(input_csv_dir, video_basename),
            os.path.join(input_csv_dir, new_file_name),
        )
        df = pd.read_csv(os.path.join(input_csv_dir, new_file_name))
        header_cols = get_number_of_header_columns_in_df(df=df)
        if header_cols == 3:
            df = df.iloc[1:]
        if file_type == "parquet":
            df = pd.read_csv(os.path.join(input_csv_dir, video_basename))
            df = df.apply(pd.to_numeric, errors="coerce")
            df.to_parquet(os.path.join(input_csv_dir, new_file_name))
            os.remove(os.path.join(input_csv_dir, video_basename))
        if file_type == "csv":
            df.to_csv(os.path.join(input_csv_dir, new_file_name), index=False)
        imported_file_paths.append(os.path.join(input_csv_dir, new_file_name))
        video_timer.stop_timer()
        print(
            f"Pose-estimation data for video {video_name} imported to SimBA project (elapsed time: {video_timer.elapsed_time_str}s)..."
        )
    return imported_file_paths


def import_single_dlc_tracking_csv_file(
    config_path: str,
    interpolation_setting: str,
    smoothing_setting: str,
    smoothing_time: int,
    file_path: str,
):
    timer = SimbaTimer(start=True)
    if (smoothing_setting == Methods.GAUSSIAN.value) or (
        smoothing_setting == Methods.SAVITZKY_GOLAY.value
    ):
        check_int(name="SMOOTHING TIME WINDOW", value=smoothing_time, min_value=1)
    check_file_exist_and_readable(file_path=file_path)
    imported_file_paths = import_dlc_csv(config_path=config_path, source=file_path)
    if interpolation_setting != "None":
        _ = Interpolate(
            input_path=imported_file_paths[0],
            config_path=config_path,
            method=interpolation_setting,
            initial_import_multi_index=True,
        )
    if (smoothing_setting == Methods.GAUSSIAN.value) or (
        smoothing_setting == Methods.SAVITZKY_GOLAY.value
    ):
        _ = Smooth(
            config_path=config_path,
            input_path=imported_file_paths[0],
            time_window=smoothing_time,
            smoothing_method=smoothing_setting,
            initial_import_multi_index=True,
        )
    timer.stop_timer()
    stdout_success(
        msg=f"Imported {str(len(imported_file_paths))} pose estimation file(s)",
        elapsed_time=timer.elapsed_time_str,
    )


def import_multiple_dlc_tracking_csv_file(
    config_path: str,
    interpolation_setting: str,
    smoothing_setting: str,
    smoothing_time: int,
    data_dir: str,
):
    timer = SimbaTimer(start=True)
    if (smoothing_setting == Methods.GAUSSIAN.value) or (
        smoothing_setting == Methods.SAVITZKY_GOLAY.value
    ):
        check_int(name="SMOOTHING TIME WINDOW", value=smoothing_time, min_value=1)
    check_if_dir_exists(in_dir=data_dir)
    imported_file_paths = import_dlc_csv(config_path=config_path, source=data_dir)
    if interpolation_setting != "None":
        _ = Interpolate(
            input_path=os.path.dirname(imported_file_paths[0]),
            config_path=config_path,
            method=interpolation_setting,
            initial_import_multi_index=True,
        )
    if (smoothing_setting == Methods.GAUSSIAN.value) or (
        smoothing_setting == Methods.SAVITZKY_GOLAY.value
    ):
        _ = Smooth(
            config_path=config_path,
            input_path=os.path.dirname(imported_file_paths[0]),
            time_window=int(smoothing_time),
            smoothing_method=smoothing_setting,
            initial_import_multi_index=True,
        )
    timer.stop_timer()
    stdout_success(
        msg=f"Imported {str(len(imported_file_paths))} pose estimation file(s)",
        elapsed_time=timer.elapsed_time_str,
    )



__author__ = "Simon Nilsson"

import json
import os

import pandas as pd

from simba.mixins.config_reader import ConfigReader
from simba.mixins.unsupervised_mixin import UnsupervisedMixin
from simba.unsupervised.enums import Clustering, Unsupervised
from simba.utils.checks import (check_file_exist_and_readable,
                                check_if_dir_exists)
from simba.utils.printing import stdout_success

CLUSTERER_PARAMETERS = "CLUSTERER HYPER-PARAMETERS"
DIMENSIONALITY_REDUCTION_PARAMETERS = "DIMENSIONALITY REDUCTION HYPER-PARAMETERS"
SCALER = "SCALER"
SCALED_DATA = "SCALED DATA"
LOW_VARIANCE_FIELDS = "LOW VARIANCE FIELDS"
FEATURE_NAMES = "FEATURE_NAMES"
FRAME_FEATURES = "FRAME_FEATURES"
FRAME_POSE = "FRAME_POSE"
FRAME_TARGETS = "FRAME_TARGETS"
BOUTS_FEATURES = "BOUTS_FEATURES"
BOUTS_TARGETS = "BOUTS_TARGETS"


class DataExtractor(UnsupervisedMixin, ConfigReader):
    """
    Extracts human-readable data from pickle holding unsupervised analyses.

    :param config_path: path to SimBA configparser.ConfigParser project_config.ini
    :param data_path: path to pickle holding unsupervised results in ``data_map.yaml`` format.
    :param data_type: The type of data to extract.
    :param settings: User-defined parameters for data extraction.

    :example:
    >>> extractor = DataExtractor(data_path='unsupervised/cluster_models/awesome_curran.pickle', data_type='BOUTS_TARGETS', settings=None, config_path='unsupervised/project_folder/project_config.ini')
    >>> extractor.run()
    """

    def __init__(
        self,
        config_path: str,
        data_path: str,
        data_type: str,
        settings: dict or None = None,
    ):
        ConfigReader.__init__(self, config_path=config_path)
        UnsupervisedMixin.__init__(self)
        if os.path.isdir(data_path):
            check_if_dir_exists(in_dir=data_path)
            self.data = self.read_pickle(data_path=data_path)
        else:
            check_file_exist_and_readable(file_path=data_path)
            self.data = {0: self.read_pickle(data_path=data_path)}
        self.settings, self.data_type = settings, data_type

    def run(self):
        if self.data_type == CLUSTERER_PARAMETERS:
            for k, v in self.data.items():
                save_path = os.path.join(
                    self.logs_path,
                    f"cluster_parameters_{v[Clustering.CLUSTER_MODEL.value][Unsupervised.HASHED_NAME.value]}.csv",
                )
                with open(save_path, "w") as fp:
                    json.dump(
                        v[Clustering.CLUSTER_MODEL.value][
                            Unsupervised.PARAMETERS.value
                        ],
                        fp,
                    )
                stdout_success(msg=f"Cluster parameters saved at {save_path}")

        elif self.data_type == DIMENSIONALITY_REDUCTION_PARAMETERS:
            for k, v in self.data.items():
                save_path = os.path.join(
                    self.logs_path,
                    f"dimensionality_reduction_parameters_{v[Unsupervised.DR_MODEL.value][Unsupervised.HASHED_NAME.value]}.csv",
                )
                with open(save_path, "w") as fp:
                    json.dump(
                        v[Unsupervised.DR_MODEL.value][Unsupervised.PARAMETERS.value],
                        fp,
                    )
                stdout_success(
                    msg=f"Dimension reduction parameters saved at {save_path}"
                )

        elif self.data_type == SCALER:
            for k, v in self.data.items():
                save_path = os.path.join(
                    self.logs_path,
                    f"scaler_{v[Unsupervised.DR_MODEL.value][Unsupervised.HASHED_NAME.value]}.pickle",
                )
                self.write_pickle(
                    data=v[Unsupervised.METHODS.value][Unsupervised.SCALER.value],
                    save_path=save_path,
                )
                stdout_success(msg=f"Scaler saved at {save_path}")

        elif self.data_type == SCALED_DATA:
            for k, v in self.data.items():
                save_path = os.path.join(
                    self.logs_path,
                    f"scaled_data_{v[Unsupervised.DR_MODEL.value][Unsupervised.HASHED_NAME.value]}.csv",
                )
                v[Unsupervised.METHODS.value][Unsupervised.SCALED_DATA.value].to_csv(
                    save_path
                )
                stdout_success(msg=f"Scaled data saved at {save_path}")

        elif self.data_type == LOW_VARIANCE_FIELDS:
            for k, v in self.data.items():
                save_path = os.path.join(
                    self.logs_path,
                    f"low_variance_fields_{v[Unsupervised.DR_MODEL.value][Unsupervised.HASHED_NAME.value]}.csv",
                )
                out_df = pd.DataFrame(
                    data=v[Unsupervised.METHODS.value][
                        Unsupervised.LOW_VARIANCE_FIELDS.value
                    ],
                    columns=["FIELD_NAMES"],
                )
                out_df.to_csv(save_path)
                stdout_success(msg=f"Low variance fields saved at {save_path}")

        elif self.data_type == FEATURE_NAMES:
            for k, v in self.data.items():
                save_path = os.path.join(
                    self.logs_path,
                    f"feature_names_{v[Unsupervised.DR_MODEL.value][Unsupervised.HASHED_NAME.value]}.csv",
                )
                out_df = pd.DataFrame(
                    data=v[Unsupervised.METHODS.value][
                        Unsupervised.FEATURE_NAMES.value
                    ],
                    columns=["FIELD_NAMES"],
                )
                out_df.to_csv(save_path)
                stdout_success(msg=f"Feature names fields saved at {save_path}")

        elif self.data_type == FRAME_FEATURES:
            for k, v in self.data.items():
                save_path = os.path.join(
                    self.logs_path,
                    f"frame_wise_features_{v[Unsupervised.DR_MODEL.value][Unsupervised.HASHED_NAME.value]}.csv",
                )
                v[Unsupervised.DATA.value][Unsupervised.FRAME_FEATURES.value].to_csv(
                    save_path
                )
                stdout_success(msg=f"Frame-wise features saved at {save_path}")

        elif self.data_type == FRAME_POSE:
            for k, v in self.data.items():
                save_path = os.path.join(
                    self.logs_path,
                    f"frame_wise_pose_estimation_{v[Unsupervised.DR_MODEL.value][Unsupervised.HASHED_NAME.value]}.csv",
                )
                v[Unsupervised.DATA.value][Unsupervised.FRAME_POSE.value].to_csv(
                    save_path
                )
                stdout_success(msg=f"Frame-wise pose saved at {save_path}")

        elif self.data_type == FRAME_TARGETS:
            for k, v in self.data.items():
                save_path = os.path.join(
                    self.logs_path,
                    f"frame_wise_target_data_{v[Unsupervised.DR_MODEL.value][Unsupervised.HASHED_NAME.value]}.csv",
                )
                v[Unsupervised.DATA.value][Unsupervised.FRAME_TARGETS.value].to_csv(
                    save_path
                )
                stdout_success(msg=f"Frame-wise target data saved at {save_path}")

        elif self.data_type == BOUTS_FEATURES:
            for k, v in self.data.items():
                save_path = os.path.join(
                    self.logs_path,
                    f"bout_features_data_{v[Unsupervised.DR_MODEL.value][Unsupervised.HASHED_NAME.value]}.csv",
                )
                v[Unsupervised.DATA.value][Unsupervised.BOUTS_FEATURES.value].to_csv(
                    save_path
                )
                stdout_success(msg=f"Bout features data saved at {save_path}")

        elif self.data_type == BOUTS_TARGETS:
            for k, v in self.data.items():
                save_path = os.path.join(
                    self.logs_path,
                    f"bout_targets_{v[Unsupervised.DR_MODEL.value][Unsupervised.HASHED_NAME.value]}.csv",
                )
                v[Unsupervised.DATA.value][Unsupervised.BOUTS_TARGETS.value].to_csv(
                    save_path
                )
                stdout_success(msg=f"Bout target data saved at {save_path}")


# test = DataExtractor(data_path='/Users/simon/Desktop/envs/troubleshooting/unsupervised/cluster_models/awesome_curran.pickle',
#                      data_type='BOUTS_TARGETS',
#                      settings=None,
#                      config_path='/Users/simon/Desktop/envs/troubleshooting/unsupervised/project_folder/project_config.ini')
#
# test.run()



__author__ = "Simon Nilsson"

import ast
import glob
import os
import re
import subprocess
from pathlib import Path
from typing import Any, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
import trafaret as t

from simba.utils.errors import (ArrayError, ColumnNotFoundError,
                                CorruptedFileError, CountError, FloatError,
                                IntegerError, InvalidFilepathError,
                                InvalidInputError, NoDataError,
                                NoFilesFoundError, NotDirectoryError,
                                ParametersFileError, StringError)
from simba.utils.warnings import NoDataFoundWarning


def check_file_exist_and_readable(file_path: Union[str, os.PathLike]) -> None:
    """
    Checks if a path points to a readable file.

    :param str file_path: Path to file on disk.
    :raise NoFilesFoundError: The file does not exist.
    :raise CorruptedFileError: The file can not be read or is zero byte size.
    """
    check_instance(
        source="FILE PATH", instance=file_path, accepted_types=(str, os.PathLike)
    )
    if not os.path.isfile(file_path):
        raise NoFilesFoundError(
            msg=f"{file_path} is not a valid file path",
            source=check_file_exist_and_readable.__name__,
        )
    elif not os.access(file_path, os.R_OK):
        raise CorruptedFileError(
            msg=f"{file_path} is not readable",
            source=check_file_exist_and_readable.__name__,
        )
    elif os.stat(file_path).st_size == 0:
        raise CorruptedFileError(
            msg=f"{file_path} is 0 bytes and contains no data.",
            source=check_file_exist_and_readable.__name__,
        )
    else:
        pass


def check_int(
    name: str,
    value: Any,
    max_value: Optional[int] = None,
    min_value: Optional[int] = None,
    raise_error: Optional[bool] = True,
) -> (bool, str):
    """
    Check if variable is a valid integer.

    :param str name: Name of variable
    :param Any value: Value of variable
    :param Optional[int] max_value: Maximum allowed value of the variable. If None, then no maximum. Default: None.
    :param Optional[int]: Minimum allowed value of the variable. If None, then no minimum. Default: None.
    :param Optional[bool] raise_error: If True, then raise error if invalid integer. Default: True.

    :return bool: False if invalid. True if valid.
    :return str: If invalid, then error msg. Else empty str.

    :examples:
    >>> check_int(name='My_fps', input=25, min_value=1)
    """
    msg = ""
    try:
        t.Int().check(value)
    except t.DataError as e:
        msg = f"{name} should be an integer number in SimBA, but is set to {str(value)}"
        if raise_error:
            raise IntegerError(msg=msg, source=check_int.__name__)
        else:
            return False, msg
    if min_value != None:
        if int(value) < min_value:
            msg = f"{name} should be MORE THAN OR EQUAL to {str(min_value)}. It is set to {str(value)}"
            if raise_error:
                raise IntegerError(msg=msg, source=check_int.__name__)
            else:
                return False, msg
    if max_value != None:
        if int(value) > max_value:
            msg = f"{name} should be LESS THAN OR EQUAL to {str(max_value)}. It is set to {str(value)}"
            if raise_error:
                raise IntegerError(msg=msg, source=check_int.__name__)
            else:
                return False, msg
    return True, msg


def check_str(
    name: str,
    value: Any,
    options: Optional[Tuple[Any]] = (),
    allow_blank: bool = False,
    raise_error: bool = True,
) -> (bool, str):
    """
    Check if variable is a valid string.

    :param str name: Name of variable
    :param Any value: Value of variable
    :param Optional[Tuple[Any]] options: Tuple of allowed strings. If empty tuple, then any string allowed. Default: ().
    :param Optional[bool] allow_blank: If True, allow empty string. Default: False.
    :param Optional[bool] raise_error: If True, then raise error if invalid string. Default: True.

    :return bool: False if invalid. True if valid.
    :return str: If invalid, then error msg. Else empty str.

    :examples:
    >>> check_str(name='split_eval', input='gini', options=['entropy', 'gini'])
    """

    msg = ""
    try:
        t.String(allow_blank=allow_blank).check(value)
    except t.DataError as e:
        msg = f"{name} should be an string in SimBA, but is set to {str(value)}"
        if raise_error:
            raise StringError(msg=msg, source=check_str.__name__)
        else:
            return False, msg
    if len(options) > 0:
        if value not in options:
            msg = f"{name} is set to {str(value)} in SimBA, but this is not a valid option: {options}"
            if raise_error:
                raise StringError(msg=msg, source=check_str.__name__)
            else:
                return False, msg
        else:
            return True, msg
    else:
        return True, msg


def check_float(
    name: str,
    value: Any,
    max_value: Optional[float] = None,
    min_value: Optional[float] = None,
    raise_error: bool = True,
) -> (bool, str):
    """
    Check if variable is a valid float.

    :param str name: Name of variable
    :param Any value: Value of variable
    :param Optional[int] max_value: Maximum allowed value of the float. If None, then no maximum. Default: None.
    :param Optional[int]: Minimum allowed value of the float. If None, then no minimum. Default: None.
    :param Optional[bool] raise_error: If True, then raise error if invalid float. Default: True.

    :return bool: False if invalid. True if valid.
    :return str: If invalid, then error msg. Else empty str.

    :examples:
    >>> check_float(name='My_float', value=0.5, max_value=1.0, min_value=0.0)
    """

    msg = ""
    try:
        t.Float().check(value)
    except t.DataError as e:
        msg = f"{name} should be a float number in SimBA, but is set to {str(value)}"
        if raise_error:
            raise FloatError(msg=msg, source=check_float.__name__)
        else:
            return False, msg
    if min_value != None:
        if float(value) < min_value:
            msg = f"{name} should be MORE THAN OR EQUAL to {str(min_value)}. It is set to {str(value)}"
            if raise_error:
                raise FloatError(msg=msg, source=check_float.__name__)
            else:
                return False, msg
    if max_value != None:
        if float(value) > max_value:
            msg = f"{name} should be LESS THAN OR EQUAL to {str(max_value)}. It is set to {str(value)}"
            if raise_error:
                raise FloatError(msg=msg, source=check_float.__name__)
            else:
                return False, msg
    return True, msg


def check_iterable_length(
    source: str,
    val: int,
    exact_accepted_length: Optional[int] = None,
    max: Optional[int] = np.inf,
    min: int = 1,
) -> None:

    if (not exact_accepted_length) and (not max) and (not min):
        raise InvalidInputError(
            msg=f"Provide exact_accepted_length or max and min values for {source}",
            source=check_iterable_length.__name__,
        )
    if exact_accepted_length:
        if val != exact_accepted_length:
            raise InvalidInputError(
                msg=f"{source} length is {val}, expected {exact_accepted_length}",
                source=check_iterable_length.__name__,
            )

    elif (val > max) or (val < min):
        raise InvalidInputError(
            msg=f"{source} value {val} does not full-fill criterion: min {min}, max{max} ",
            source=check_iterable_length.__name__,
        )


def check_instance(
    source: str, instance: object, accepted_types: Union[Tuple[object], object]
) -> None:
    """
    Check if an instance is an acceptable type.

    :param str name: Arbitrary name of instance used for interpretable error msg. Can also be the name of the method.
    :param object instance: A data object.
    :param Union[Tuple[object], object] accepted_types: Accepted instance types. E.g., (Polygon, pd.DataFrame) or Polygon.
    """

    if not isinstance(instance, accepted_types):
        raise InvalidInputError(
            msg=f"{source} requires {accepted_types}, got {type(instance)}",
            source=source,
        )


def get_fn_ext(filepath: Union[os.PathLike, str]) -> (str, str, str):
    """
    Split file path into three components: (i) directory, (ii) file name, and (iii) file extension.

    :parameter str filepath: Path to file.
    :return str: File directory name
    :return str: File name
    :return str: File extension

    :example:
    >>> get_fn_ext(filepath='C:/My_videos/MyVideo.mp4')
    >>> ('My_videos', 'MyVideo', '.mp4')
    """
    file_extension = Path(filepath).suffix
    try:
        file_name = os.path.basename(filepath.rsplit(file_extension, 1)[0])
    except ValueError:
        raise InvalidFilepathError(
            msg=f"{filepath} is not a valid filepath", source=get_fn_ext.__name__
        )
    dir_name = os.path.dirname(filepath)
    return dir_name, file_name, file_extension


def check_if_filepath_list_is_empty(filepaths: List[str], error_msg: str) -> None:
    """
    Check if a list is empty

    :param List[str]: List of file-paths.
    :raise NoFilesFoundError: The list is empty.
    """

    if len(filepaths) == 0:
        raise NoFilesFoundError(
            msg=error_msg, source=check_if_filepath_list_is_empty.__name__
        )
    else:
        pass


def check_all_file_names_are_represented_in_video_log(
    video_info_df: pd.DataFrame, data_paths: List[Union[str, os.PathLike]]
) -> None:
    """
    Helper to check that all files are represented in a dataframe of the SimBA `project_folder/logs/video_info.csv`
    file.

    :param pd.DataFrame video_info_df: List of file-paths.
    :param List[Union[str, os.PathLike]] data_paths: List of file-paths.
    :raise ParametersFileError: The list is empty.
    """

    missing_videos = []
    for file_path in data_paths:
        video_name = get_fn_ext(file_path)[1]
        if video_name not in list(video_info_df["Video"]):
            missing_videos.append(video_name)
    if len(missing_videos) > 0:
        raise ParametersFileError(
            msg=f"SimBA could not find {len(missing_videos)} video(s) in the video_info.csv file. Make sure all videos analyzed are represented in the project_folder/logs/video_info.csv file. MISSING VIDEOS: {missing_videos}"
        )


def check_if_dir_exists(in_dir: Union[str, os.PathLike]) -> None:
    """
    Check if a directory path exists.

    :param str in_dir: Putative directory path.
    :raise NotDirectoryError: The directory does not exist.
    """
    if not os.path.isdir(in_dir):
        raise NotDirectoryError(
            msg=f"{in_dir} is not a valid directory",
            source=check_if_dir_exists.__name__,
        )


def check_that_column_exist(
    df: pd.DataFrame, column_name: Union[str, os.PathLike, List[str]], file_name: str
) -> None:
    """
    Check if single named field or a list of fields exist within a dataframe.

    :param pd.DataFrame df:
    :param str column_name: Name or names of field(s).
    :param str file_name: Path of ``df`` on disk.
    :raise ColumnNotFoundError: The ``column_name`` does not exist within ``df``.
    """

    if type(column_name) == str:
        column_name = [column_name]
    for column in column_name:
        if column not in df.columns:
            raise ColumnNotFoundError(
                column_name=column,
                file_name=file_name,
                source=check_that_column_exist.__name__,
            )


def check_if_valid_input(
    name: str, input: str, options: List[str], raise_error: bool = True
) -> (bool, str):
    """
    Check if string variable is valid option

    :param str name: Atrbitrary name of variable.
    :param Any input: Value of variable.
    :param List[str] options: Allowed options of ``input``
    :param Optional[bool] raise_error: If True, then raise error if invalid value. Default: True.

    :return bool: False if invalid. True if valid.
    :return str: If invalid, then error msg. Else, empty str.

    :example:
    >>> check_if_valid_input(name='split_eval', input='gini', options=['entropy', 'gini'])
    >>> (True, '')
    """

    msg = ""
    if input not in options:
        msg = f"{name} is set to {str(input)}, which is an invalid setting. OPTIONS {options}"
        if raise_error:
            raise InvalidInputError(msg=msg, source=check_if_valid_input.__name__)
        else:
            return False, msg
    else:
        return True, msg


def check_minimum_roll_windows(
    roll_windows_values: List[int], minimum_fps: float
) -> List[int]:
    """
    Remove any rolling temporal window that are shorter than a single frame in
    any of the videos within the project.

    :param List[int] roll_windows_values: Rolling temporal windows represented as frame counts. E.g., [10, 15, 30, 60]
    :param float minimum_fps: The lowest fps of the videos that are to be analyzed. E.g., 10.

    :return List[int]: roll_windows_values without impassable windows.
    """

    for win in range(len(roll_windows_values)):
        if minimum_fps < roll_windows_values[win]:
            roll_windows_values[win] = minimum_fps
        else:
            pass
    roll_windows_values = list(set(roll_windows_values))
    return roll_windows_values


def check_same_number_of_rows_in_dfs(dfs: List[pd.DataFrame]) -> bool:
    """
    Helper to check that each dataframe in list contains an equal number of rows

    :param List[pd.DataFrame] dfs: List of dataframes.
    :return bool: True if dataframes has an equal number of rows. Else False.

    >>> df_1, df_2 = pd.DataFrame([[1, 2], [1, 2]]), pd.DataFrame([[4, 2], [9, 3], [1, 5]])
    >>> check_same_number_of_rows_in_dfs(dfs=[df_1, df_2])
    >>> False
    >>> df_1, df_2 = pd.DataFrame([[1, 2], [1, 2]]), pd.DataFrame([[4, 2], [9, 3]])
    >>> True
    """

    row_cnt = None
    for df_cnt, df in enumerate(dfs):
        if df_cnt == 0:
            row_cnt = len(df)
        else:
            if len(df) != row_cnt:
                return False
    return True


def check_if_headers_in_dfs_are_unique(dfs: List[pd.DataFrame]) -> List[str]:
    """
    Helper to check heaaders in multiple dataframes are unique.

    :param List[pd.DataFrame] dfs: List of dataframes.
    :return List[str]: List of columns headers seen in multiple dataframes. Empty if None.

    :examples:
    >>> df_1, df_2 = pd.DataFrame([[1, 2]], columns=['My_column_1', 'My_column_2']), pd.DataFrame([[4, 2]], columns=['My_column_3', 'My_column_1'])
    >>> check_if_headers_in_dfs_are_unique(dfs=[df_1, df_2])
    >>> ['My_column_1']
    """
    seen_headers = []
    for df_cnt, df in enumerate(dfs):
        seen_headers.extend(list(df.columns))
    duplicates = list(set([x for x in seen_headers if seen_headers.count(x) > 1]))
    return duplicates


def check_if_string_value_is_valid_video_timestamp(value: str, name: str) -> None:
    """
    Helper to check if a string is in a valid HH:MM:SS format

    :param str value: Timestamp in HH:MM:SS format.
    :param str name: An arbitrary string name of the timestamp.
    :raises InvalidInputError: If the timestamp is in invalid format

    :example:
    >>> check_if_string_value_is_valid_video_timestamp(value='00:0b:10', name='My time stamp')
    >>> "InvalidInputError: My time stamp is should be in the format XX:XX:XX where X is an integer between 0-9"
    >>> check_if_string_value_is_valid_video_timestamp(value='00:00:10', name='My time stamp'
    """
    r = re.compile(r"^\d{2}:\d{2}:\d{2}(\.\d+)?$")
    if not r.match(value):
        raise InvalidInputError(
            msg=f"{name} should be in the format XX:XX:XX:XXXX or XX:XX:XX where X is an integer between 0-9. Got: {value}",
            source=check_if_string_value_is_valid_video_timestamp.__name__,
        )
    else:
        pass


def check_that_hhmmss_start_is_before_end(
    start_time: str, end_time: str, name: str
) -> None:
    """
    Helper to check that a start time in HH:MM:SS or HH:MM:SS:MS format is before an end time in HH:MM:SS or HH:MM:SS:MS format

    :param str start_time: Period start time in HH:MM:SS format.
    :param str end_time: Period end time in HH:MM:SS format.
    :param int name: Name of the variable
    :raises InvalidInputError: If end time is before the start time.

    :example:
    >>> check_that_hhmmss_start_is_before_end(start_time='00:00:05', end_time='00:00:01', name='My time period')
    >>> "InvalidInputError: My time period has an end-time which is before the start-time"
    >>> check_that_hhmmss_start_is_before_end(start_time='00:00:01', end_time='00:00:05')
    """

    if len(start_time.split(":")) != 3:
        raise InvalidInputError(
            f"Invalid time-stamp: ({start_time}). HH:MM:SS or HH:MM:SS.MS format required"
        )
    elif len(end_time.split(":")) != 3:
        raise InvalidInputError(
            f"Invalid time-stamp: ({end_time}). HH:MM:SS or HH:MM:SS.MS format required"
        )
    start_h, start_m, start_s = start_time.split(":")
    end_h, end_m, end_s = end_time.split(":")
    start_val = int(start_h) * 3600 + int(start_m) * 60 + float(start_s)
    end_val = int(end_h) * 3600 + int(end_m) * 60 + float(end_s)
    if end_val < start_val:
        raise InvalidInputError(
            f"{name} has an end-time which is before the start-time.",
            source=check_that_hhmmss_start_is_before_end.__name__,
        )


def check_nvidea_gpu_available() -> bool:
    """
    Helper to check of NVIDEA GPU is available via ``nvidia-smi``.
    returns bool: True if nvidia-smi returns not None. Else False.
    """
    try:
        subprocess.check_output("nvidia-smi")
        return True
    except Exception:
        return False


def check_ffmpeg_available() -> bool:
    """
    Helper to check of FFMpeg is available via subprocess ``ffmpeg``.

    returns bool: True if ``ffmpeg`` returns not None. Else False.
    """
    try:
        subprocess.call("ffmpeg", stderr=subprocess.DEVNULL, stdout=subprocess.DEVNULL)
        return True
    except Exception:
        return False


def check_if_valid_rgb_str(
    input: str,
    delimiter: str = ",",
    return_cleaned_rgb_tuple: bool = True,
    reverse_returned: bool = True,
):
    """
    Helper to check if a string is a valid representation of an RGB color.

    :param str input: Value to check as string. E.g., '(166, 29, 12)' or '22,32,999'
    :param str delimiter: The delimiter between subsequent values in the rgb input string.
    :param bool return_cleaned_rgb_tuple: If True, and input is a valid rgb, then returns a "clean" rgb tuple: Eg. '166, 29, 12' -> (166, 29, 12). Else, returns None.
    :param bool reverse_returned: If True and return_cleaned_rgb_tuple is True, reverses to returned cleaned rgb tuple (e.g., RGB becomes BGR) before returning it.

    :example:
    >>> check_if_valid_rgb_str(input='(50, 25, 100)', return_cleaned_rgb_tuple=True, reverse_returned=True)
    >>> (100, 25, 50)
    """

    input = input.replace(" ", "")
    if input.count(delimiter) != 2:
        raise InvalidInputError(msg=f"{input} in not a valid RGB color")
    values = input.split(",")
    rgb = []
    for value in values:
        val = "".join(c for c in value if c.isdigit())
        check_int(
            name="RGB value", value=val, max_value=255, min_value=0, raise_error=True
        )
        rgb.append(val)
    rgb = tuple([int(x) for x in rgb])

    if return_cleaned_rgb_tuple:
        if reverse_returned:
            rgb = rgb[::-1]
        return rgb


def check_if_valid_rgb_tuple(data: Tuple[int, int, int]) -> bool:
    check_instance(
        source=check_if_valid_rgb_tuple.__name__, instance=data, accepted_types=tuple
    )
    check_iterable_length(
        source=check_if_valid_rgb_tuple.__name__, val=len(data), exact_accepted_length=3
    )
    for i in range(len(data)):
        check_int(
            name="RGB value",
            value=data[i],
            max_value=255,
            min_value=0,
            raise_error=True,
        )


def check_if_list_contains_values(
    data: List[Union[float, int, str]],
    values: List[Union[float, int, str]],
    name: str,
    raise_error: bool = True,
) -> None:
    """
    Helper to check if values are represeted in a list. E.g., make sure annotatations of behvaior absent and present are represented in annitation column

    :param List[Union[float, int, str]] data: List of values. E.g., annotation column represented as list.
    :param List[Union[float, int, str]] values: Values to conform present. E.g., [0, 1].
    :param str name: Arbitrary name of the data for more useful error msg.
    :param bool raise_error: If True, raise error of not all values can be found in data. Else, print warning.

    :example:
    >>> check_if_list_contains_values(data=[1,2, 3, 4, 0], values=[0, 1, 6], name='My_data')
    """

    data, missing_values = list(set(data)), []
    for value in values:
        if value not in data:
            missing_values.append(value)

    if len(missing_values) > 0 and raise_error:
        raise NoDataError(
            msg=f"{name} does not contain the following expected values: {missing_values}",
            source=check_if_list_contains_values.__name__,
        )

    elif len(missing_values) > 0 and not raise_error:
        NoDataFoundWarning(
            msg=f"{name} does not contain the following expected values: {missing_values}",
            source=check_if_list_contains_values.__name__,
        )


def check_valid_hex_color(color_hex: str, raise_error: Optional[bool] = True) -> bool:
    """
    Check if given string represents a valid hexadecimal color code.

    :param str color_hex: A string representing a hexadecimal color code, either in the format '#RRGGBB' or '#RGB'.
    :param bool raise_error: If True, raise an exception when the color_hex is invalid; if False, return False instead. Default is True.
    :return bool: True if the color_hex is a valid hexadecimal color code; False otherwise (if raise_error is False).
    :raises IntegerError: If the color_hex is an invalid hexadecimal color code and raise_error is True.
    """

    hex_regex = re.compile(r"^#([0-9a-fA-F]{6}|[0-9a-fA-F]{3})$")
    match = hex_regex.match(color_hex)
    if match is None and raise_error:
        raise IntegerError(
            msg=f"{color_hex} is an invalid hex color",
            source=check_valid_hex_color.__name__,
        )
    elif match is None and not raise_error:
        return False
    else:
        return True


def check_if_2d_array_has_min_unique_values(data: np.ndarray, min: int) -> bool:
    """
    Check if a 2D NumPy array has at least a minimum number of unique rows.

    For example, use when creating shapely Polygons or Linestrings, which typically requires at least 2 or three unique
    body-part coordinates.

    :param np.ndarray data: Input 2D array to be checked.
    :param np.ndarray min: Minimum number of unique rows required.
    :return bool: True if the input array has at least the specified minimum number of unique rows, False otherwise.

    :example:
    >>> data = np.array([[0, 0], [0, 0], [0, 0], [0, 1]])
    >>> check_if_2d_array_has_min_unique_values(data=data, min=2)
    >>> True
    """

    if len(data.shape) != 2:
        raise CountError(
            msg=f"Requires input array of two dimensions, found {data.size}",
            source=check_if_2d_array_has_min_unique_values.__name__,
        )
    sliced_data = np.unique(data, axis=0)
    if sliced_data.shape[0] < min:
        return False
    else:
        return True


def check_if_module_has_import(parsed_file: ast.Module, import_name: str) -> bool:
    """
    Check if a Python module has a specific import statement.

    Used for e.g., user custom feature extraction classes in ``simba.utils.custom_feature_extractor.CustomFeatureExtractor``.

    :parameter ast.Module file_path: The abstract syntax tree (AST) of the Python module.
    :parameter str import_name: The name of the module or package to check for in the import statements.
    :parameter bool: True if the specified import is found in the module, False otherwise.

    :example:
    >>> parsed_file = ast.parse(Path('/simba/misc/piotr.py').read_text())
    >>> check_if_module_has_import(parsed_file=parsed_file, import_name='argparse')
    >>> True
    """
    imports = [
        n for n in parsed_file.body if isinstance(n, (ast.Import, ast.ImportFrom))
    ]
    for i in imports:
        for name in i.names:
            if name.name == import_name:
                return True
    return False


def check_if_valid_img(
    data: np.ndarray, source: Optional[str] = "", raise_error: Optional[bool] = True
) -> Union[bool, None]:
    """
    Check if a variable is a valid image.

    :parameter str source: Name of the variable and/or class origin for informative error messaging and logging.
    :parameter np.ndarray data: Data variable to check if a valid image representation.
    :parameter Optional[bool] raise_error: If True, raise InvalidInputError if invalid image representation. Else, return bool.
    """

    check_instance(
        source=check_if_valid_img.__name__, instance=data, accepted_types=np.ndarray
    )
    if (data.ndim != 2) and (data.ndim != 3):
        if raise_error:
            raise InvalidInputError(
                msg=f"The {source} data is not a valid image. It has {data.ndim} dimensions",
                source=check_if_valid_img.__name__,
            )
        else:
            return False
    if data.dtype not in [np.uint8, np.uint16, np.float32, np.float64]:
        if raise_error:
            raise InvalidInputError(
                msg=f"The {source} data is not a valid image. It is dtype {data.dtype}",
                source=check_if_valid_img.__name__,
            )
        else:
            return False


def check_that_dir_has_list_of_filenames(
    dir: Union[str, os.PathLike],
    file_name_lst: List[str],
    file_type: Optional[str] = "csv",
):
    """
    Check that all file names in a list has an equivalent file in a specified directory. E.g., check if all files in the outlier corrected folder has an equivalent file in the featurues_extracted directory.

    :example:
    >>> file_name_lst = glob.glob('/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/csv/outlier_corrected_movement' + '/*.csv')
    >>> check_that_dir_has_list_of_filenames(dir = '/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/csv/features_extracted', file_name_lst=file_name_lst)
    """

    files_in_dir = glob.glob(dir + f"/*.{file_type}")
    files_in_dir = [os.path.basename(x) for x in files_in_dir]
    for file_name in file_name_lst:
        if os.path.basename(file_name) not in files_in_dir:
            raise NoFilesFoundError(
                msg=f"File name {os.path.basename(file_name)} could not be found in the directory {dir}",
                source=check_that_dir_has_list_of_filenames.__name__,
            )


def check_valid_array(
    data: np.ndarray,
    source: Optional[str] = "",
    accepted_ndims: Optional[List[Tuple[int]]] = None,
    accepted_sizes: Optional[List[int]] = None,
    accepted_dtypes: Optional[List[str]] = None,
    accepted_shapes: Optional[List[Tuple[int]]] = None,
) -> None:
    """
    Check if the given  array satisfies specified criteria regarding its dimensions, shape, and data type.

    :parameter np.ndarray data: The numpy array to be checked.
    :parameter Optional[str] source: A string identifying the source, name, or purpose of the array for interpretable error messaging.
    :parameter Optional[Tuple[int]] accepted_ndims: List of tuples representing acceptable dimensions. If provided, checks whether the array's number of dimensions matches any tuple in the list.
    :parameter Optional[List[int]] accepted_sizes: List of acceptable sizes for the array's shape. If provided, checks whether the length of the array's shape matches any value in the list.
    :parameter Optional[List[str]] accepted_dtypes: List of acceptable data types for the array. If provided, checks whether the array's data type matches any string in the list.

    :example:
    >>> data = np.array([[1, 2], [3, 4]])
    >>> check_valid_array(data, source="Example", accepted_ndims=(4, 3), accepted_sizes=[2], accepted_dtypes=['int'])
    """

    check_instance(source=source, instance=data, accepted_types=np.ndarray)
    if accepted_ndims is not None:
        if data.ndim not in accepted_ndims:
            raise ArrayError(
                msg=f"Array not of acceptable dimensions. Found {data.ndim}, accepted: {accepted_ndims}: {source}",
                source=check_valid_array.__name__,
            )
    if accepted_sizes is not None:
        if len(data.shape) not in accepted_sizes:
            raise ArrayError(
                msg=f"Array not of acceptable size. Found {len(data.shape)}, accepted: {accepted_sizes}: {source}",
                source=check_valid_array.__name__,
            )
    if accepted_dtypes is not None:
        if data.dtype not in accepted_dtypes:
            raise ArrayError(
                msg=f"Array not of acceptable type. Found {data.dtype}, accepted: {accepted_dtypes}: {source}",
                source=check_valid_array.__name__,
            )
    if accepted_shapes is not None:
        if data.shape not in accepted_shapes:
            raise ArrayError(
                msg=f"Array not of acceptable shape. Found {data.shape}, accepted: {accepted_shapes}: {source}",
                source=check_valid_array.__name__,
            )


def check_valid_lst(
    data: list,
    source: Optional[str] = "",
    valid_dtypes: Optional[Tuple[Any]] = None,
    min_len: Optional[int] = 1,
    max_len: Optional[int] = None,
    raise_error: Optional[bool] = True,
) -> bool:
    """
    Check the validity of a list based on passed  criteria.

    :parameter list data: The input list to be validated.
    :parameter Optional[str] source: A string indicating the source or context of the data for informative error messaging.
    :parameter Optional[Tuple[Any]] valid_dtypes: A tuple of accepted data types. If provided, check if all elements in the list have data types in this tuple.
    :parameter Optional[int] min_len: The minimum allowed length of the list.
    :parameter Optional[int] max_len: The maximum allowed length of the list.
    :parameter Optional[bool] raise_error: If True, raise an InvalidInputError if any validation fails. If False, return False instead of raising an error.
    :return bool: True if all validation criteria are met, False otherwise.

    :example:
    >>> check_valid_lst(data=[1, 2, 'three'], valid_dtypes=(int, str), min_len=2, max_len=5)
    >>> check_valid_lst(data=[1, 2, 3], valid_dtypes=(int,), min_len=3)
    """

    check_instance(source=source, instance=data, accepted_types=list)
    if valid_dtypes is not None:
        for dtype in set([type(x) for x in data]):
            if dtype not in valid_dtypes:
                if raise_error:
                    raise InvalidInputError(
                        msg=f"Invalid data type found in list. Found {dtype}, accepted: {valid_dtypes}",
                        source=source,
                    )
                else:
                    return False
    if min_len is not None:
        if len(data) < min_len:
            if raise_error:
                raise InvalidInputError(
                    msg=f"Invalid length of list. Found {len(data)}, minimum accepted: {min_len}",
                    source=source,
                )
            else:
                return False
    if max_len is not None:
        if len(data) > max_len:
            if raise_error:
                raise InvalidInputError(
                    msg=f"Invalid length of list. Found {len(data)}, maximum accepted: {min_len}",
                    source=source,
                )
            else:
                return False

    return True



__author__ = "Simon Nilsson"

try:
    from typing import Literal
except:
    from typing_extensions import Literal
import numpy as np
from numba import jit, prange

from simba.mixins.feature_extraction_mixin import FeatureExtractionMixin


class FeatureExtractionSupplemental(FeatureExtractionMixin):
    """
    Additional feature extraction method not called by default feature extraction classes from ``simba.feature_extractors``.

    """

    def __init__(self):
        FeatureExtractionMixin.__init__(self)

    @staticmethod
    @jit(nopython=True)
    def _helper_euclidean_distance_timeseries_change(
        distances: np.ndarray, time_windows: np.ndarray, fps: int
    ):
        """
        Private jitted helper called by ``simba.mixins.feature_extraction_supplemental_mixin.FeatureExtractionSupplemental.euclidean_distance_timeseries_change``
        """
        results = np.full((distances.shape[0], time_windows.shape[0]), np.nan)
        for window_cnt in prange(time_windows.shape[0]):
            frms = int(time_windows[window_cnt] * fps)
            shifted_distances = np.copy(distances)
            shifted_distances[0:frms] = np.nan
            shifted_distances[frms:] = distances[:-frms]
            shifted_distances[np.isnan(shifted_distances)] = distances[
                np.isnan(shifted_distances)
            ]
            results[:, window_cnt] = distances - shifted_distances

        return results

    def euclidean_distance_timeseries_change(
        self,
        location_1: np.ndarray,
        location_2: np.ndarray,
        fps: int,
        px_per_mm: float,
        time_windows: np.ndarray = np.array([0.2, 0.4, 0.8, 1.6]),
    ) -> np.ndarray:
        """
        Compute the difference in distance between two points in the current frame versus N.N seconds ago. E.g.,
        computes if two points are traveling away from each other (positive output values) or towards each other
        (negative output values) relative to reference time-point(s)

        .. image:: _static/img/euclid_distance_change.png
           :width: 700
           :align: center

        :parameter ndarray location_1: 2D array of size len(frames) x 2 representing pose-estimated locations of body-part one
        :parameter ndarray location_2: 2D array of size len(frames) x 2 representing pose-estimated locations of body-part two
        :parameter int fps: Fps of the recorded video.
        :parameter float px_per_mm: The pixels per millimeter in the video.
        :parameter np.ndarray time_windows: Time windows to compare.
        :return np.array: Array of size location_1.shape[0] x time_windows.shape[0]

        :example:
        >>> location_1 = np.random.randint(low=0, high=100, size=(2000, 2)).astype('float32')
        >>> location_2 = np.random.randint(low=0, high=100, size=(2000, 2)).astype('float32')
        >>> distances = self.euclidean_distance_timeseries_change(location_1=location_1, location_2=location_2, fps=10, px_per_mm=4.33, time_windows=np.array([0.2, 0.4, 0.8, 1.6]))
        """
        distances = self.framewise_euclidean_distance(
            location_1=location_1, location_2=location_2, px_per_mm=px_per_mm
        )
        return self._helper_euclidean_distance_timeseries_change(
            distances=distances, fps=fps, time_windows=time_windows
        ).astype(int)

    @staticmethod
    @jit(nopython=True)
    def peak_ratio(data: np.ndarray, bin_size_s: int, fps: int):
        """
        Compute the ratio of peak values relative to number of values within each seqential
        time-period represented of ``bin_size_s`` seconds. Peak is defined as value is higher than
        in the prior observation (i.e., no future data is involved in comparison).

        :parameter ndarray data: 1D array of size len(frames) representing feature values.
        :parameter int bin_size_s: The size of the buckets in seconds.
        :parameter int fps: Frame-rate of recorded video.
        :return np.ndarray: Array of size data.shape[0] with peak counts as ratio of len(frames).

        .. image:: _static/img/peak_cnt.png
           :width: 700
           :align: center

        :example:
        >>> data = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
        >>> FeatureExtractionSupplemental().peak_ratio(data=data, bin_size_s=1, fps=10)
        >>> [0.9 0.9 0.9 0.9 0.9 0.9 0.9 0.9 0.9 0.9]
        >>> data = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])
        >>> FeatureExtractionSupplemental().peak_ratio(data=data, bin_size_s=1, fps=10)
        >>> [0.9 0.9 0.9 0.9 0.9 0.9 0.9 0.9 0.9 0.9 0.  0.  0.  0.  0.  0.  0.  0. 0.  0. ]
        """

        window_size, results = int(bin_size_s * fps), np.full((data.shape[0]), -1.0)
        data = np.split(data, list(range(window_size, data.shape[0], window_size)))
        start, end = 0, data[0].shape[0]
        for i in prange(len(data)):
            peak_cnt = 0
            if data[i][0] > data[i][1]:
                peak_cnt += 1
            if data[i][-1] > data[i][-2]:
                peak_cnt += 1
            for j in prange(1, len(data[i]) - 1):
                if data[i][j] > data[i][j - 1]:
                    peak_cnt += 1
            peak_ratio = peak_cnt / data[i].shape[0]
            results[start:end] = peak_ratio
            start, end = start + len(data[i]), end + len(data[i])
        return results

    @staticmethod
    @jit(nopython=True)
    def rolling_peak_count_ratio(data: np.ndarray, time_windows: np.ndarray, fps: int):

        results = np.full((data.shape[0], time_windows.shape[0]), -1.0)
        for i in prange(time_windows.shape[0]):
            window_size = int(time_windows[i] * fps)
            for j in prange(window_size, data.shape[0]):
                window_data = data[j - window_size : j]
                peak_cnt = 0
                if window_data[0] > window_data[1]:
                    peak_cnt += 1
                if window_data[-1] > window_data[-2]:
                    peak_cnt += 1
                for k in prange(1, len(window_data) - 1):
                    if window_data[j] > window_data[j - 1]:
                        peak_cnt += 1
                peak_ratio = peak_cnt / window_data.shape[0]
                results[j, i] = peak_ratio
        print(results)

    @staticmethod
    @jit(nopython=True)
    def rolling_categorical_switches_ratio(
        data: np.ndarray, time_windows: np.ndarray, fps: int
    ) -> np.ndarray:
        """
        Compute the ratio of in categorical feature switches within rolling windows.

        .. attention::
           Output for initial frames where [current_frm - window_size] < 0, are populated with ``0``.

        .. image:: _static/img/feature_switches.png
           :width: 700
           :align: center

        :parameter np.ndarray data: 1d array of feature values
        :parameter np.ndarray time_windows: Rolling time-windows as floats in seconds. E.g., [0.2, 0.4, 0.6]
        :parameter int fps: fps of the recorded video
        :returns np.ndarray: Size data.shape[0] x time_windows.shape[0] array

        :example:
        >>> data = np.array([0, 1, 1, 1, 4, 5, 6, 7, 8, 9])
        >>> FeatureExtractionSupplemental().rolling_categorical_switches_ratio(data=data, time_windows=np.array([1.0]), fps=10)
        >>> [[-1][-1][-1][-1][-1][-1][-1][-1][-1][ 0.7]]
        >>> data = np.array(['A', 'B', 'B', 'B', 'C', 'D', 'E', 'F', 'G', 'H'])
        >>> FeatureExtractionSupplemental().rolling_categorical_switches_ratio(data=data, time_windows=np.array([1.0]), fps=10)
        >>> [[-1][-1][-1][-1][-1][-1][-1][-1][-1][ 0.7]]
        """

        results = np.full((data.shape[0], time_windows.shape[0]), -1.0)
        for time_window in prange(time_windows.shape[0]):
            jump_frms = int(time_windows[time_window] * fps)
            for current_frm in prange(jump_frms, data.shape[0] + 1):
                time_slice = data[current_frm - jump_frms : current_frm]
                current_value, unique_cnt = time_slice[0], 0
                for i in prange(1, time_slice.shape[0]):
                    if time_slice[i] != current_value:
                        unique_cnt += 1
                    current_value = time_slice[i]
                print(unique_cnt, time_slice.shape[0])
                results[current_frm - 1][time_window] = unique_cnt / time_slice.shape[0]
        return results

    @staticmethod
    @jit(nopython=True)
    def consecutive_time_series_categories_count(data: np.ndarray, fps: int):
        """
        Compute the count of consecutive milliseconds the feature value has remained static. For example,
        compute for how long in milleseconds the animal has remained in the current cardinal direction or the
        within an ROI.

        .. image:: _static/img/categorical_consecitive_time.png
           :width: 700
           :align: center

        :parameter np.ndarray data: 1d array of feature values
        :parameter int fps: Frame-rate of video.
        :returns np.ndarray: Array of size data.shape[0]

        :example:
        >>> data = np.array([0, 1, 1, 1, 4, 5, 6, 7, 8, 9])
        >>> FeatureExtractionSupplemental().consecutive_time_series_categories_count(data=data, fps=10)
        >>> [0.1, 0.1, 0.2, 0.3, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1]
        >>> data = np.array(['A', 'B', 'B', 'B', 'C', 'D', 'E', 'F', 'G', 'H'])
        >>> [0.1, 0.1, 0.2, 0.3, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1]
        """

        results = np.full((data.shape[0]), 0.0)
        results[0] = 1
        for i in prange(1, data.shape[0]):
            if data[i] == data[i - 1]:
                results[i] = results[i - 1] + 1
            else:
                results[i] = 1

        return results / fps

    @staticmethod
    @jit(nopython=True)
    def rolling_horizontal_vs_vertical_movement(
        data: np.ndarray, pixels_per_mm: float, time_windows: np.ndarray, fps: int
    ) -> np.ndarray:
        """
        Compute the movement along the x-axis relative to the y-axis in rolling time bins.

        .. attention::
           Output for initial frames where [current_frm - window_size] < 0, are populated with ``0``.

        .. image:: _static/img/x_vs_y_movement.png
           :width: 700
           :align: center

        :parameter np.ndarray data: 2d array of size len(frames)x2 with body-part coordinates.
        :parameter int fps: FPS of the recorded video
        :parameter float pixels_per_mm: Pixels per millimeter of recorded video.
        :returns np.ndarray: Size data.shape[0] x time_windows.shape[0] array
        :parameter np.ndarray time_windows: Rolling time-windows as floats in seconds. E.g., [0.2, 0.4, 0.6]
        :returns np.ndarray: Size data.shape[0] x time_windows.shape[0]. Greater values denote greater movement on x-axis relative to y-axis.

        :example:
        >>> data = np.array([[250, 250], [250, 250], [250, 250], [250, 500], [500, 500], 500, 500]]).astype(float)
        >>> FeatureExtractionSupplemental().rolling_horizontal_vs_vertical_movement(data=data, time_windows=np.array([1.0]), fps=2, pixels_per_mm=1)
        >>> [[  -1.][   0.][   0.][-250.][ 250.][   0.]]
        """

        results = np.full((data.shape[0], time_windows.shape[0]), 0)
        for time_window in prange(time_windows.shape[0]):
            jump_frms = int(time_windows[time_window] * fps)
            for current_frm in prange(jump_frms, results.shape[0] + 1):
                x_movement = (
                    np.sum(
                        np.abs(
                            np.ediff1d(data[current_frm - jump_frms : current_frm, 0])
                        )
                    )
                    / pixels_per_mm
                )
                y_movement = (
                    np.sum(
                        np.abs(
                            np.ediff1d(data[current_frm - jump_frms : current_frm, 1])
                        )
                    )
                    / pixels_per_mm
                )
                results[current_frm - 1][time_window] = x_movement - y_movement

        return results

    @staticmethod
    @jit(nopython=True)
    def border_distances(
        data: np.ndarray,
        pixels_per_mm: float,
        img_resolution: np.ndarray,
        time_window: float,
        fps: int,
    ):
        """
        Compute the mean distance of key-point to the left, right, top, and bottom sides of the image in
        rolling time-windows. Uses a straight line.

        .. image:: _static/img/border_distance.png
           :width: 700
           :align: center

        .. attention::
           Output for initial frames where [current_frm - window_size] < 0 will be populated with ``-1``.

        :parameter np.ndarray data: 2d array of size len(frames)x2 with body-part coordinates.
        :parameter np.ndarray img_resolution: Resolution of video in WxH format.
        :parameter float pixels_per_mm: Pixels per millimeter of recorded video.
        :parameter int fps: FPS of the recorded video
        :parameter float time_windows: Rolling time-window as floats in seconds. E.g., ``0.2``
        :returns np.ndarray: Size data.shape[0] x time_windows.shape[0] array with millimeter distances from LEFT, RIGH, TOP, BOTTOM,

        :example:
        >>> data = np.array([[250, 250], [250, 250], [250, 250], [500, 500],[500, 500], [500, 500]]).astype(float)
        >>> img_resolution = np.array([500, 500])
        >>> FeatureExtractionSupplemental().border_distances(data=data, img_resolution=img_resolution, time_window=1, fps=2, pixels_per_mm=1)
        >>> [[-1, -1, -1, -1][250, 250, 250, 250][250, 250, 250, 250][375, 125, 375, 125][500, 0, 500, 0][500, 0, 500, 0]]
        """

        results = np.full((data.shape[0], 4), -1.0)
        window_size = int(time_window * fps)
        for current_frm in prange(window_size, results.shape[0] + 1):
            distances = np.full((4, window_size, 1), np.nan)
            windowed_locs = data[current_frm - window_size : current_frm]
            for bp_cnt, bp_loc in enumerate(windowed_locs):
                distances[0, bp_cnt] = np.linalg.norm(
                    np.array([0, bp_loc[1]]) - bp_loc
                )  # left
                distances[1, bp_cnt] = np.linalg.norm(
                    np.array([img_resolution[0], bp_loc[1]]) - bp_loc
                )  # right
                distances[2, bp_cnt] = np.linalg.norm(
                    np.array([bp_loc[0], 0]) - bp_loc
                )  # top
                distances[3, bp_cnt] = np.linalg.norm(
                    np.array([bp_loc[0], img_resolution[1]]) - bp_loc
                )  # bottpm
            for i in prange(4):
                results[current_frm - 1][i] = np.mean(distances[i]) / pixels_per_mm

        return results.astype(np.int32)


# import time
#
# start = time.time()
# for i in range(10000):
#     results = FeatureExtractionSupplemental().brunner_munzel(sample_1=sample_1, sample_2=sample_2)
# print(time.time() - start)
# #
# start = time.time()
# for i in range(10000):
#     bm = stats.brunnermunzel(x=sample_1, y=sample_2)
# print(time.time() - start)
#
#
#
# # print(results)
# # print(bm)



__author__ = "Simon Nilsson"

import os
from tkinter import *
from tkinter import messagebox
from typing import Dict, List, Optional, Tuple, Union

import PIL.Image
from PIL import ImageTk

from simba.mixins.config_reader import ConfigReader
from simba.pose_importers import trk_importer
from simba.pose_importers.dlc_importer_csv import (
    import_multiple_dlc_tracking_csv_file, import_single_dlc_tracking_csv_file)
from simba.pose_importers.import_mars import MarsImporter
from simba.pose_importers.madlc_importer import MADLCImporterH5
from simba.pose_importers.read_DANNCE_mat import (import_DANNCE_file,
                                                  import_DANNCE_folder)
from simba.pose_importers.sleap_csv_importer import SLEAPImporterCSV
from simba.pose_importers.sleap_h5_importer import SLEAPImporterH5
from simba.pose_importers.sleap_slp_importer import SLEAPImporterSLP
from simba.ui.tkinter_functions import (DropDownMenu, Entry_Box, FileSelect,
                                        FolderSelect, hxtScrollbar)
from simba.utils.checks import (check_file_exist_and_readable, check_float,
                                check_if_dir_exists, check_int, check_str)
from simba.utils.enums import ConfigKey, Formats, Options
from simba.utils.errors import CountError, NoFilesFoundError
from simba.utils.lookups import (get_color_dict, get_icons_paths,
                                 get_named_colors)
from simba.utils.read_write import (copy_multiple_videos_to_project,
                                    copy_single_video_to_project,
                                    find_core_cnt, read_config_entry,
                                    read_config_file)


class PopUpMixin(object):
    """
    Methods for pop-up windows in SimBA. E.g., common methods for creating pop-up windows with drop-downs,
    checkboxes, entry-boxes, listboxes etc.

    :param str title: Pop-up window title
    :param Optional[configparser.Configparser] config_path: path to SimBA project_config.ini. If path, the project config is read in. If None, the project config is not read in.
    :param tuple size: HxW of the pop-up window. The size of the pop-up window in pixels.
    :param bool main_scrollbar: If True, the pop-up window is scrollable.
    """

    def __init__(
        self,
        title: str,
        config_path: Optional[str] = None,
        main_scrollbar: Optional[bool] = True,
        size: Tuple[int, int] = (960, 720),
    ):

        # self.main_frm = Toplevel()
        # self.main_frm.minsize(size[0], size[1])
        # self.main_frm.wm_title(title)
        # self.main_frm.lift()
        # self.main_frm = Canvas(hxtScrollbar(self.main_frm))
        self.root = Toplevel()
        self.root.minsize(size[0], size[1])
        self.root.wm_title(title)
        self.root.lift()
        if main_scrollbar:
            self.main_frm = Canvas(hxtScrollbar(self.root))
        else:
            self.main_frm = Canvas(self.root)
        self.main_frm.pack(fill="both", expand=True)
        self.palette_options = Options.PALETTE_OPTIONS.value
        self.resolutions = Options.RESOLUTION_OPTIONS.value
        self.shading_options = Options.HEATMAP_SHADING_OPTIONS.value
        self.heatmap_bin_size_options = Options.HEATMAP_BIN_SIZE_OPTIONS.value
        self.dpi_options = Options.DPI_OPTIONS.value
        self.colors = get_named_colors()
        self.colors_dict = get_color_dict()
        self.cpu_cnt, _ = find_core_cnt()
        self.menu_icons = get_icons_paths()
        for k in self.menu_icons.keys():
            self.menu_icons[k]["img"] = ImageTk.PhotoImage(
                image=PIL.Image.open(
                    os.path.join(
                        os.path.dirname(__file__), self.menu_icons[k]["icon_path"]
                    )
                )
            )
        if config_path:
            ConfigReader.__init__(self, config_path=config_path, read_video_info=False)

    def create_clf_checkboxes(
        self,
        main_frm: Frame,
        clfs: List[str],
        title: str = "SELECT CLASSIFIER ANNOTATIONS",
    ):
        """
        Creates a labelframe with one checkbox per classifier, and inserts the labelframe into the bottom of the pop-up window.

        .. note::
           Legacy. Use ``create_cb_frame`` instead.

        """

        self.choose_clf_frm = LabelFrame(
            self.main_frm, text=title, font=Formats.LABELFRAME_HEADER_FORMAT.value
        )
        self.clf_selections = {}
        for clf_cnt, clf in enumerate(clfs):
            self.clf_selections[clf] = BooleanVar(value=False)
            self.calculate_distance_moved_cb = Checkbutton(
                self.choose_clf_frm, text=clf, variable=self.clf_selections[clf]
            )
            self.calculate_distance_moved_cb.grid(row=clf_cnt, column=0, sticky=NW)
        self.choose_clf_frm.grid(row=self.children_cnt_main(), column=0, sticky=NW)

    def create_cb_frame(
        self,
        main_frm: Frame,
        cb_titles: List[str],
        frm_title: str,
        command: Optional[object] = None,
    ) -> Dict[str, BooleanVar]:
        """
        Creates a labelframe with one checkbox per classifier, and inserts the labelframe into the bottom of the pop-up window.

        :param Frame main_frm: The tkinter pop-up window.
        :param List[str] cb_titles: List of strings representing the names of the checkboxes.
        :param str frm_title: Title of the frame.
        :return Dict[str, BooleanVar]: Dictionary holding the ``cb_titles`` as keys and the BooleanVar representing if the checkbox is ticked or not.
        """

        cb_frm = LabelFrame(
            main_frm, text=frm_title, font=Formats.LABELFRAME_HEADER_FORMAT.value
        )
        cb_dict = {}
        for cnt, title in enumerate(cb_titles):
            cb_dict[title] = BooleanVar(value=False)
            if command is not None:
                if isinstance(command, object):
                    cb = Checkbutton(
                        cb_frm,
                        text=title,
                        variable=cb_dict[title],
                        command=lambda k=cb_titles[cnt]: command(k),
                    )
            else:
                cb = Checkbutton(cb_frm, text=title, variable=cb_dict[title])
            cb.grid(row=cnt, column=0, sticky=NW)
        cb_frm.grid(row=self.children_cnt_main(), column=0, sticky=NW)
        return cb_dict

    def frame_of_radiobuttons(
        self,
        main_frm: Frame,
        titles: List[str],
        frm_title: str,
        command: Optional[object] = None,
        default_idx: Optional[int] = 0,
    ):

        selection_var = StringVar()
        radiobuttons_frm = LabelFrame(
            main_frm, text=frm_title, font=Formats.LABELFRAME_HEADER_FORMAT.value
        )
        for cnt, title in enumerate(titles):
            radio_button = Radiobutton(
                main_frm,
                text=titles[cnt],
                variable=selection_var,
                value=titles[cnt],
                command=command,
            )
            radio_button.grid(row=cnt, column=0, sticky=NW)
        radiobuttons_frm.grid(row=self.children_cnt_main(), column=0, sticky=NW)
        selection_var.set(value=titles[default_idx])
        return selection_var

    def place_frm_at_top_right(self, frm: Toplevel):
        """
        Place a TopLevel tkinter pop-up at the top right of the monitor. Note: call before putting scrollbars or converting to Canvas.
        """
        screen_width, screen_height = frm.winfo_screenwidth(), frm.winfo_screenheight()
        window_width, window_height = frm.winfo_width(), frm.winfo_height()
        x_position = screen_width - window_width
        frm.geometry(f"{window_width}x{window_height}+{x_position}+{0}")

    def create_dropdown_frame(
        self,
        main_frm: Frame,
        drop_down_titles: List[str],
        drop_down_options: List[str],
        frm_title: str,
    ) -> Dict[str, DropDownMenu]:
        """
        Creates a labelframe with dropdown menus and inserts it at the bottom of the pop-up window.

        :param Frame main_frm: The tkinter pop-up window.
        :param List[str] drop_down_titles: The dropdown menu names
        :param List[str] drop_down_options: The options in each dropdown. All dropdowns must have the same options.
        :param str frm_title: Title of the frame.
        :return Dict[str, BooleanVar]: Dictionary holding the ``drop_down_titles`` and the drop-down menus as values.

        """

        dropdown_frm = LabelFrame(
            main_frm, text=frm_title, font=Formats.LABELFRAME_HEADER_FORMAT.value
        )
        dropdown_dict = {}
        for cnt, title in enumerate(drop_down_titles):
            dropdown_dict[title] = DropDownMenu(
                dropdown_frm, title, drop_down_options, "35"
            )
            dropdown_dict[title].setChoices(drop_down_options[0])
            dropdown_dict[title].grid(row=cnt, column=0, sticky=NW)
        dropdown_frm.grid(row=self.children_cnt_main(), column=0, sticky=NW)
        return dropdown_dict

    def create_choose_animal_cnt_dropdown(self):
        if hasattr(self, "animal_cnt_frm"):
            self.animal_cnt_frm.destroy()
        animal_cnt_options = set(range(1, self.project_animal_cnt + 1))
        self.animal_cnt_frm = LabelFrame(
            self.main_frm,
            text="SELECT NUMBER OF ANIMALS",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
        )
        self.animal_cnt_dropdown = DropDownMenu(
            self.animal_cnt_frm, "# of animals", animal_cnt_options, "12"
        )
        self.animal_cnt_dropdown.setChoices(max(animal_cnt_options))
        self.animal_cnt_confirm_btn = Button(
            self.animal_cnt_frm,
            text="Confirm",
            command=lambda: self.update_body_parts(),
        )
        self.animal_cnt_frm.grid(row=self.children_cnt_main(), sticky=NW)
        self.animal_cnt_dropdown.grid(row=self.children_cnt_main(), column=1, sticky=NW)
        self.animal_cnt_confirm_btn.grid(
            row=self.children_cnt_main(), column=2, sticky=NW
        )
        self.create_choose_body_parts_frm()
        self.update_body_parts()

    def create_choose_body_parts_frm(self):
        if hasattr(self, "body_part_frm"):
            self.body_part_frm.destroy()
        self.body_parts_dropdown_dict = {}
        self.body_part_frm = LabelFrame(
            self.main_frm,
            text="CHOOSE ANIMAL BODY-PARTS",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            name="choose animal body-parts",
        )
        self.body_part_frm.grid(row=self.children_cnt_main(), sticky=NW)

    def update_body_parts(self):
        for child in self.body_part_frm.winfo_children():
            child.destroy()
        for animal_cnt in range(int(self.animal_cnt_dropdown.getChoices())):
            animal_name = list(self.animal_bp_dict.keys())[animal_cnt]
            self.body_parts_dropdown_dict[animal_name] = DropDownMenu(
                self.body_part_frm,
                f"{animal_name} body-part:",
                self.body_parts_lst,
                "25",
            )
            self.body_parts_dropdown_dict[animal_name].grid(
                row=animal_cnt, column=1, sticky=NW
            )
            self.body_parts_dropdown_dict[animal_name].setChoices(
                self.body_parts_lst[animal_cnt]
            )

    def create_time_bin_entry(self):
        if hasattr(self, "time_bin_frm"):
            self.time_bin_frm.destroy()
        self.time_bin_frm = LabelFrame(
            self.main_frm, text="TIME BIN", font=Formats.LABELFRAME_HEADER_FORMAT.value
        )
        self.time_bin_entrybox = Entry_Box(
            self.time_bin_frm, "Time-bin size (s): ", "12", validation="numeric"
        )
        self.time_bin_entrybox.grid(row=0, column=0, sticky=NW)
        self.time_bin_frm.grid(row=self.children_cnt_main(), column=0, sticky=NW)

    def create_run_frm(self, run_function: object, title: str = "RUN") -> None:
        """
        Create a label frame with a single button with a specified callback.

        :param object run_function: The function/method callback of the button.
        :param str title: The title of the frame.
        """

        if hasattr(self, "run_frm"):
            self.run_frm.destroy()
            self.run_btn.destroy()
        self.run_frm = LabelFrame(
            self.main_frm,
            text="RUN",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="black",
        )
        self.run_btn = Button(
            self.run_frm, text=title, fg="blue", command=lambda: run_function()
        )
        self.run_frm.grid(row=self.children_cnt_main() + 1, column=0, sticky=NW)
        self.run_btn.grid(row=0, column=0, sticky=NW)

    def create_choose_number_of_body_parts_frm(
        self, project_body_parts: List[str], run_function: object
    ):
        """
        Many menus depend on how many animals the user choose to compute metrics for. Thus, we need to populate the menus
        dynamically. This function creates a single drop-down menu where the user select the number of animals the
        user choose to compute metrics for. It inserts this drop-down iat the bottom of the pop-up window, and ties this dropdown menu
        choice to a callback.

        :param List[str] project_body_parts: Options of the dropdown menu.
        :param object run_function: Function tied to the choice in the dropdown menu.
        """

        self.bp_cnt_frm = LabelFrame(
            self.main_frm,
            text="SELECT NUMBER OF BODY-PARTS",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
        )
        self.bp_cnt_dropdown = DropDownMenu(
            self.bp_cnt_frm,
            "# of body-parts",
            list(range(1, len(project_body_parts) + 1)),
            "12",
        )
        self.bp_cnt_dropdown.setChoices(1)
        self.bp_cnt_confirm_btn = Button(
            self.bp_cnt_frm,
            text="Confirm",
            command=lambda: self.create_choose_bp_frm(project_body_parts, run_function),
        )
        self.bp_cnt_frm.grid(row=0, sticky=NW)
        self.bp_cnt_dropdown.grid(row=0, column=0, sticky=NW)
        self.bp_cnt_confirm_btn.grid(row=0, column=1, sticky=NW)

    def add_to_listbox_from_entrybox(self, list_box: Listbox, entry_box: Entry_Box):
        """
        Add a value that populates a tkinter entry_box to a tkinter listbox.

        :param Listbox list_box: The tkinter Listbox to add the value to.
        :param Entry_Box entry_box: The tkinter Entry_Box containing the value that should be added to the list_box.
        """

        value = entry_box.entry_get
        check_float(name="VALUE", value=value)
        list_box_content = [float(x) for x in list_box.get(0, END)]
        if float(value) not in list_box_content:
            list_box.insert(0, value)

    def add_value_to_listbox(self, list_box: Listbox, value: float):
        """
        Add a float value to a tkinter listbox.

        :param Listbox list_box: The tkinter Listbox to add the value to.
        :param float value: Value to add to the listbox.
        """
        list_box.insert(0, value)

    def add_values_to_several_listboxes(
        self, list_boxes: List[Listbox], values: List[float]
    ):
        """
        Add N values to N listboxes. E.g., values[0] will be added to list_boxes[0].

        :param List[Listbox] list_boxes: List of Listboxes that the values should be added to.
        :param List[float] values: List of floats that will be added to the list_boxes.
        """

        if len(list_boxes) != len(values):
            raise CountError(
                msg="Value count and list-boxes count are not equal",
                source=self.__class__.__name__,
            )
        for i in range(len(list_boxes)):
            list_boxes[i].insert(0, values[i])

    def remove_from_listbox(self, list_box: Listbox):
        """
        Remove the current selection in a listbox from a listbox.

        :param Listbox list_box: The listbox that the current selection should be removed from.
        """

        selection = list_box.curselection()
        if selection:
            list_box.delete(selection[0])

    def update_file_select_box_from_dropdown(
        self, filename: str, fileselectbox: FileSelect
    ):
        """
        Updates the text inside a tkinter FileSelect entrybox with a new string.
        """
        fileselectbox.filePath.set(filename)

    def check_if_selected_video_path_exist_in_project(
        self, video_path: Union[str, os.PathLike]
    ):
        if not os.path.isfile(video_path):
            raise NoFilesFoundError(
                msg=f"Selected video {os.path.basename(video_path)} is not a video file in the SimBA project video directory."
            )

    def create_choose_bp_frm(self, project_body_parts, run_function):
        if hasattr(self, "body_part_frm"):
            self.body_part_frm.destroy()
        self.body_parts_dropdowns = {}
        self.body_part_frm = LabelFrame(
            self.main_frm,
            text="CHOOSE ANIMAL BODY-PARTS",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            name="choose animal body-parts",
        )
        self.body_part_frm.grid(row=self.children_cnt_main(), sticky=NW)
        for bp_cnt in range(int(self.bp_cnt_dropdown.getChoices())):
            self.body_parts_dropdowns[bp_cnt] = DropDownMenu(
                self.body_part_frm,
                f"Body-part {str(bp_cnt+1)}:",
                project_body_parts,
                "25",
            )
            self.body_parts_dropdowns[bp_cnt].grid(row=bp_cnt, column=0, sticky=NW)
            self.body_parts_dropdowns[bp_cnt].setChoices(project_body_parts[bp_cnt])
        self.create_run_frm(run_function=run_function)

    def choose_bp_frm(self, parent: LabelFrame, bp_options: list):
        self.body_parts_dropdowns = {}
        self.body_part_frm = LabelFrame(
            parent,
            text="CHOOSE ANIMAL BODY-PARTS",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            name="choose animal body-parts",
        )
        self.body_part_frm.grid(row=self.frame_children(frame=parent), sticky=NW)
        for bp_cnt in range(int(self.animal_cnt_dropdown.getChoices())):
            self.body_parts_dropdowns[bp_cnt] = DropDownMenu(
                self.body_part_frm, f"Body-part {str(bp_cnt + 1)}:", bp_options, "20"
            )
            self.body_parts_dropdowns[bp_cnt].grid(row=bp_cnt, column=0, sticky=NW)
            self.body_parts_dropdowns[bp_cnt].setChoices(bp_options[bp_cnt])

    def children_cnt_main(self) -> int:
        """
        Find the number of children (e.g., labelframes) currently exist within a main pop-up window. Useful for finding the
        row at which a new frame within the window should be inserted.
        """

        return int(len(list(self.main_frm.children.keys())))

    def frame_children(self, frame: Frame) -> int:
        """
        Find the number of children (e.g., labelframes) currently exist within specified frame.Similar to ``children_cnt_main``,
        but accepts a specific frame rather than the main frame beeing hardcoded.
        """

        return int(len(list(frame.children.keys())))

    def update_config(self) -> None:
        """Helper to update the SimBA project config file"""
        with open(self.config_path, "w") as f:
            self.config.write(f)

    def show_smoothing_entry_box_from_dropdown(self, choice: str):
        if choice == "None":
            self.smoothing_time_eb.grid_forget()
        if (choice == "Gaussian") or (choice == "Savitzky Golay"):
            self.smoothing_time_eb.grid(row=0, column=1, sticky=E)

    def choose_bp_threshold_frm(self, parent: LabelFrame):
        self.probability_frm = LabelFrame(
            parent,
            text="PROBABILITY THRESHOLD",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
        )
        self.probability_frm.grid(
            row=self.frame_children(frame=parent), column=0, sticky=NW
        )
        self.probability_entry = Entry_Box(
            self.probability_frm, "Probability threshold: ", labelwidth=20
        )
        self.probability_entry.entry_set("0.00")
        self.probability_entry.grid(row=0, column=0, sticky=NW)

    def enable_dropdown_from_checkbox(
        self, check_box_var: BooleanVar, dropdown_menus: List[DropDownMenu]
    ):
        """
        Given a single checkbox, enable a bunch of dropdowns if the checkbox is ticked, and disable the dropdowns if
        the checkbox is un-ticked.

        :param BooleanVar check_box_var: The checkbox associated tkinter BooleanVar.
        :param List[DropDownMenu] dropdown_menus: List of dropdowns which status is controlled by the ``check_box_var``.
        """

        if check_box_var.get():
            for menu in dropdown_menus:
                menu.enable()
        else:
            for menu in dropdown_menus:
                menu.disable()

    def create_entry_boxes_from_entrybox(
        self, count: int, parent: Frame, current_entries: list
    ):
        check_int(name="CLASSIFIER COUNT", value=count, min_value=1)
        for entry in current_entries:
            entry.destroy()
        for clf_cnt in range(int(count)):
            entry = Entry_Box(parent, f"Classifier {str(clf_cnt+1)}:", labelwidth=15)
            current_entries.append(entry)
            entry.grid(row=clf_cnt + 2, column=0, sticky=NW)

    def create_animal_names_entry_boxes(self, animal_cnt: str):
        check_int(name="NUMBER OF ANIMALS", value=animal_cnt, min_value=0)
        if hasattr(self, "animal_names_frm"):
            self.animal_names_frm.destroy()
        if not hasattr(self, "multi_animal_id_list"):
            self.multi_animal_id_list = []
            for i in range(int(animal_cnt)):
                self.multi_animal_id_list.append(f"Animal {i+1}")
        self.animal_names_frm = Frame(self.animal_settings_frm, pady=5, padx=5)
        self.animal_name_entry_boxes = {}
        for i in range(int(animal_cnt)):
            self.animal_name_entry_boxes[i + 1] = Entry_Box(
                self.animal_names_frm, f"Animal {str(i+1)} name: ", "25"
            )
            if i <= len(self.multi_animal_id_list) - 1:
                self.animal_name_entry_boxes[i + 1].entry_set(
                    self.multi_animal_id_list[i]
                )
            self.animal_name_entry_boxes[i + 1].grid(row=i, column=0, sticky=NW)

        self.animal_names_frm.grid(row=1, column=0, sticky=NW)

    def enable_entrybox_from_checkbox(
        self,
        check_box_var: BooleanVar,
        entry_boxes: List[Entry_Box],
        reverse: bool = False,
    ):
        """
        Given a single checkbox, enable or disable a bunch of entry-boxes based on the status of the checkbox.

        :param BooleanVar check_box_var: The checkbox associated tkinter BooleanVar.
        :param List[Entry_Box] entry_boxes: List of entry-boxes which status is controlled by the ``check_box_var``.
        :param bool reverse: If False, the entry-boxes are enabled with the checkbox is ticked. Else, the entry-boxes are enabled if checkbox is unticked. Default: False.
        """

        if reverse:
            if check_box_var.get():
                for box in entry_boxes:
                    box.set_state("disable")
            else:
                for box in entry_boxes:
                    box.set_state("normal")
        else:
            if check_box_var.get():
                for box in entry_boxes:
                    box.set_state("normal")
            else:
                for box in entry_boxes:
                    box.set_state("disable")

    def create_import_pose_menu(
        self, parent_frm: Frame, idx_row: int = 0, idx_column: int = 0
    ):

        def run_call(
            data_type: str,
            interpolation: str,
            smoothing: str,
            smoothing_window: str,
            animal_names: dict,
            data_path: str,
            tracking_data_type: str or None = None,
        ):

            smooth_settings = {}
            smooth_settings["Method"] = smoothing
            smooth_settings["Parameters"] = {}
            smooth_settings["Parameters"]["Time_window"] = smoothing_window

            if smooth_settings["Method"] != "None":
                check_int(
                    name="SMOOTHING TIME WINDOW", value=smoothing_window, min_value=1
                )

            if self.animal_name_entry_boxes is None:
                raise CountError(
                    msg="Select animal number and animal names BEFORE importing data.",
                    source=self.__class__.__name__,
                )

            animal_ids = []
            if len(list(animal_names.items())) == 1:
                animal_ids.append("Animal_1")
            else:
                for animal_cnt, animal_entry_box in animal_names.items():
                    check_str(
                        name=f"ANIMAL {str(animal_cnt)} NAME",
                        value=animal_entry_box.entry_get,
                        allow_blank=False,
                    )
                    animal_ids.append(animal_entry_box.entry_get)

            self.config = read_config_file(config_path=self.config_path)
            self.config.set(
                ConfigKey.MULTI_ANIMAL_ID_SETTING.value,
                ConfigKey.MULTI_ANIMAL_IDS.value,
                ",".join(animal_ids),
            )
            self.update_config()

            if data_type == "H5 (multi-animal DLC)":
                dlc_multi_animal_importer = MADLCImporterH5(
                    config_path=self.config_path,
                    data_folder=data_path,
                    file_type=tracking_data_type,
                    id_lst=animal_ids,
                    interpolation_settings=interpolation,
                    smoothing_settings=smooth_settings,
                )
                dlc_multi_animal_importer.run()

            if data_type == "SLP (SLEAP)":
                sleap_importer = SLEAPImporterSLP(
                    project_path=self.config_path,
                    data_folder=data_path,
                    id_lst=animal_ids,
                    interpolation_settings=interpolation,
                    smoothing_settings=smooth_settings,
                )
                sleap_importer.run()

            if data_type == "TRK (multi-animal APT)":
                try:
                    trk_importer(
                        self.config_path,
                        data_path,
                        animal_ids,
                        interpolation,
                        smooth_settings,
                    )
                except Exception as e:
                    messagebox.showerror("Error", str(e))

            if data_type == "CSV (SLEAP)":
                sleap_csv_importer = SLEAPImporterCSV(
                    config_path=self.config_path,
                    data_folder=data_path,
                    id_lst=animal_ids,
                    interpolation_settings=interpolation,
                    smoothing_settings=smooth_settings,
                )
                sleap_csv_importer.run()

            if data_type == "H5 (SLEAP)":
                sleap_h5_importer = SLEAPImporterH5(
                    config_path=self.config_path,
                    data_folder=data_path,
                    id_lst=animal_ids,
                    interpolation_settings=interpolation,
                    smoothing_settings=smooth_settings,
                )
                sleap_h5_importer.run()

        def import_menu(data_type_choice: str):
            if hasattr(self, "choice_frm"):
                self.choice_frm.destroy()
            self.choice_frm = Frame(self.import_tracking_frm)
            self.animal_name_entry_boxes = None

            self.interpolation_frm = LabelFrame(
                self.choice_frm, text="INTERPOLATION METHOD", pady=5, padx=5
            )
            self.interpolation_dropdown = DropDownMenu(
                self.interpolation_frm,
                "Interpolation method: ",
                Options.INTERPOLATION_OPTIONS_W_NONE.value,
                "25",
            )
            self.interpolation_dropdown.setChoices("None")
            self.interpolation_frm.grid(row=0, column=0, sticky=NW)
            self.interpolation_dropdown.grid(row=0, column=0, sticky=NW)

            self.smoothing_frm = LabelFrame(
                self.choice_frm, text="SMOOTHING METHOD", pady=5, padx=5
            )
            self.smoothing_dropdown = DropDownMenu(
                self.smoothing_frm,
                "Smoothing",
                Options.SMOOTHING_OPTIONS_W_NONE.value,
                "25",
                com=self.show_smoothing_entry_box_from_dropdown,
            )
            self.smoothing_dropdown.setChoices("None")
            self.smoothing_time_eb = Entry_Box(
                self.smoothing_frm,
                "Period (ms):",
                labelwidth="12",
                width=10,
                validation="numeric",
            )
            self.smoothing_frm.grid(row=1, column=0, sticky=NW)
            self.smoothing_dropdown.grid(row=0, column=0, sticky=NW)

            if data_type_choice in [
                "CSV (DLC/DeepPoseKit)",
                "MAT (DANNCE 3D)",
                "JSON (BENTO)",
            ]:
                if data_type_choice == "CSV (DLC/DeepPoseKit)":
                    self.import_directory_frm = LabelFrame(
                        self.choice_frm, text="IMPORT DLC CSV DIRECTORY", pady=5, padx=5
                    )
                    self.import_directory_select = FolderSelect(
                        self.import_directory_frm, "Input DIRECTORY:", lblwidth=25
                    )
                    self.import_dir_btn = Button(
                        self.import_directory_frm,
                        fg="blue",
                        text="Import DIRECTORY to SimBA project",
                        command=lambda: import_multiple_dlc_tracking_csv_file(
                            config_path=self.config_path,
                            interpolation_setting=self.interpolation_dropdown.getChoices(),
                            smoothing_setting=self.smoothing_dropdown.getChoices(),
                            smoothing_time=self.smoothing_time_eb.entry_get,
                            data_dir=self.import_directory_select.folder_path,
                        ),
                    )

                    self.import_single_frm = LabelFrame(
                        self.choice_frm, text="IMPORT DLC CSV FILE", pady=5, padx=5
                    )
                    self.import_file_select = FileSelect(
                        self.import_single_frm,
                        "Input FILE:",
                        lblwidth=25,
                        file_types=[("CSV", "*.csv")],
                    )
                    self.import_file_btn = Button(
                        self.import_single_frm,
                        fg="blue",
                        text="Import FILE to SimBA project",
                        command=lambda: import_single_dlc_tracking_csv_file(
                            config_path=self.config_path,
                            interpolation_setting=self.interpolation_dropdown.getChoices(),
                            smoothing_setting=self.smoothing_dropdown.getChoices(),
                            smoothing_time=self.smoothing_time_eb.entry_get,
                            file_path=self.import_file_select.file_path,
                        ),
                    )

                elif data_type_choice == "MAT (DANNCE 3D)":
                    self.import_directory_frm = LabelFrame(
                        self.choice_frm,
                        text="IMPORT DANNCE MAT DIRECTORY",
                        pady=5,
                        padx=5,
                    )
                    self.import_directory_select = FolderSelect(
                        self.import_directory_frm, "Input DIRECTORY:", lblwidth=25
                    )
                    self.import_dir_btn = Button(
                        self.import_directory_frm,
                        fg="blue",
                        text="Import directory to SimBA project",
                        command=lambda: import_DANNCE_folder(
                            config_path=self.config_path,
                            folder_path=self.import_directory_select.folder_path,
                            interpolation_method=self.interpolation_dropdown.getChoices(),
                        ),
                    )

                    self.import_single_frm = LabelFrame(
                        self.choice_frm, text="IMPORT DANNCE CSV FILE", pady=5, padx=5
                    )
                    self.import_file_select = FileSelect(
                        self.import_single_frm, "Input FILE:", lblwidth=25
                    )
                    self.import_file_btn = Button(
                        self.import_single_frm,
                        fg="blue",
                        text="Import file to SimBA project",
                        command=lambda: import_DANNCE_file(
                            config_path=self.config_path,
                            file_path=self.import_file_select.file_path,
                            interpolation_method=self.interpolation_dropdown.getChoices(),
                        ),
                    )

                elif data_type_choice == "JSON (BENTO)":
                    self.import_directory_frm = LabelFrame(
                        self.choice_frm,
                        text="IMPORT MARS JSON DIRECTORY",
                        pady=5,
                        padx=5,
                    )
                    self.import_directory_select = FolderSelect(
                        self.import_directory_frm, "Input DIRECTORY:", lblwidth=25
                    )
                    self.import_dir_btn = Button(
                        self.import_directory_frm,
                        fg="blue",
                        text="Import directory to SimBA project",
                        command=lambda: MarsImporter(
                            config_path=self.config_path,
                            data_path=self.import_directory_select.folder_path,
                            interpolation_method=self.interpolation_dropdown.getChoices(),
                            smoothing_method={
                                "Method": self.smoothing_dropdown.getChoices(),
                                "Parameters": {
                                    "Time_window": self.smoothing_time_eb.entry_get
                                },
                            },
                        ),
                    )

                    self.import_single_frm = LabelFrame(
                        self.choice_frm, text="IMPORT MARS JSON FILE", pady=5, padx=5
                    )
                    self.import_file_select = FileSelect(
                        self.import_single_frm, "Input FILE:", lblwidth=25
                    )
                    self.import_file_btn = Button(
                        self.import_single_frm,
                        fg="blue",
                        text="Import file to SimBA project",
                        command=lambda: MarsImporter(
                            config_path=self.config_path,
                            data_path=self.import_directory_select.folder_path,
                            interpolation_method=self.interpolation_dropdown.getChoices(),
                            smoothing_method={
                                "Method": self.smoothing_dropdown.getChoices(),
                                "Parameters": {
                                    "Time_window": self.smoothing_time_eb.entry_get
                                },
                            },
                        ),
                    )
                self.import_directory_frm.grid(row=2, column=0, sticky=NW)
                self.import_directory_select.grid(row=0, column=0, sticky=NW)
                self.import_dir_btn.grid(row=1, column=0, sticky=NW)

                self.import_single_frm.grid(row=3, column=0, sticky=NW)
                self.import_file_select.grid(row=0, column=0, sticky=NW)
                self.import_file_btn.grid(row=1, column=0, sticky=NW)

            elif data_type_choice in [
                "SLP (SLEAP)",
                "H5 (multi-animal DLC)",
                "TRK (multi-animal APT)",
                "CSV (SLEAP)",
                "H5 (SLEAP)",
            ]:
                self.animal_settings_frm = LabelFrame(
                    self.choice_frm, text="ANIMAL SETTINGS", pady=5, padx=5
                )
                animal_cnt_entry_box = Entry_Box(
                    self.animal_settings_frm,
                    "ANIMAL COUNT:",
                    "25",
                    validation="numeric",
                )
                animal_cnt_entry_box.entry_set(val=self.project_animal_cnt)
                animal_cnt_confirm = Button(
                    self.animal_settings_frm,
                    text="CONFIRM",
                    fg="blue",
                    command=lambda: self.create_animal_names_entry_boxes(
                        animal_cnt=animal_cnt_entry_box.entry_get
                    ),
                )
                self.create_animal_names_entry_boxes(
                    animal_cnt=animal_cnt_entry_box.entry_get
                )
                self.animal_settings_frm.grid(row=4, column=0, sticky=NW)
                animal_cnt_entry_box.grid(row=0, column=0, sticky=NW)
                animal_cnt_confirm.grid(row=0, column=1, sticky=NW)

                self.data_dir_frm = LabelFrame(
                    self.choice_frm, text="DATA DIRECTORY", pady=5, padx=5
                )
                self.import_frm = LabelFrame(
                    self.choice_frm, text="IMPORT", pady=5, padx=5
                )

                if data_type_choice == "H5 (multi-animal DLC)":
                    self.tracking_type_frm = LabelFrame(
                        self.choice_frm, text="TRACKING DATA TYPE", pady=5, padx=5
                    )
                    self.dlc_data_type_option_dropdown = DropDownMenu(
                        self.tracking_type_frm,
                        "Tracking type",
                        Options.MULTI_DLC_TYPE_IMPORT_OPTION.value,
                        labelwidth=25,
                    )
                    self.dlc_data_type_option_dropdown.setChoices(
                        Options.MULTI_DLC_TYPE_IMPORT_OPTION.value[1]
                    )
                    self.tracking_type_frm.grid(row=5, column=0, sticky=NW)
                    self.dlc_data_type_option_dropdown.grid(row=0, column=0, sticky=NW)

                    self.data_dir_select = FolderSelect(
                        self.data_dir_frm, "H5 DLC DIRECTORY: ", lblwidth=25
                    )
                    self.instructions_lbl = Label(
                        self.data_dir_frm,
                        text="Please import videos before importing the \n multi animal DLC tracking data",
                    )
                    self.run_btn = Button(
                        self.import_frm,
                        text="IMPORT DLC .H5",
                        fg="blue",
                        command=lambda: run_call(
                            data_type=data_type_choice,
                            interpolation=self.interpolation_dropdown.getChoices(),
                            smoothing=self.smoothing_dropdown.getChoices(),
                            smoothing_window=self.smoothing_time_eb.entry_get,
                            animal_names=self.animal_name_entry_boxes,
                            data_path=self.data_dir_select.folder_path,
                            tracking_data_type=self.dlc_data_type_option_dropdown.getChoices(),
                        ),
                    )

                elif data_type_choice == "SLP (SLEAP)":
                    self.data_dir_select = FolderSelect(
                        self.data_dir_frm, "SLP SLEAP DIRECTORY: ", lblwidth=25
                    )
                    self.instructions_lbl = Label(
                        self.data_dir_frm,
                        text="Please import videos before importing the \n multi animal SLEAP tracking data if you are tracking more than ONE animal",
                    )
                    self.run_btn = Button(
                        self.import_frm,
                        text="IMPORT SLEAP .SLP",
                        fg="blue",
                        command=lambda: run_call(
                            data_type=data_type_choice,
                            interpolation=self.interpolation_dropdown.getChoices(),
                            smoothing=self.smoothing_dropdown.getChoices(),
                            smoothing_window=self.smoothing_time_eb.entry_get,
                            animal_names=self.animal_name_entry_boxes,
                            data_path=self.data_dir_select.folder_path,
                        ),
                    )

                elif data_type_choice == "TRK (multi-animal APT)":
                    self.data_dir_select = FolderSelect(
                        self.data_dir_frm, "TRK APT DIRECTORY: ", lblwidth=25
                    )
                    self.instructions_lbl = Label(
                        self.data_dir_frm,
                        text="Please import videos before importing the \n multi animal TRK tracking data",
                    )
                    self.run_btn = Button(
                        self.import_frm,
                        text="IMPORT APT .TRK",
                        fg="blue",
                        command=lambda: run_call(
                            data_type=data_type_choice,
                            interpolation=self.interpolation_dropdown.getChoices(),
                            smoothing=self.smoothing_dropdown.getChoices(),
                            smoothing_window=self.smoothing_time_eb.entry_get,
                            animal_names=self.animal_name_entry_boxes,
                            data_path=self.data_dir_select.folder_path,
                        ),
                    )

                elif data_type_choice == "CSV (SLEAP)":
                    self.data_dir_select = FolderSelect(
                        self.data_dir_frm, "CSV SLEAP DIRECTORY:", lblwidth=25
                    )
                    self.instructions_lbl = Label(
                        self.data_dir_frm,
                        text="Please import videos before importing the sleap csv tracking data \n if you are tracking more than ONE animal",
                    )
                    self.run_btn = Button(
                        self.import_frm,
                        text="IMPORT SLEAP .CSV",
                        fg="blue",
                        command=lambda: run_call(
                            data_type=data_type_choice,
                            interpolation=self.interpolation_dropdown.getChoices(),
                            smoothing=self.smoothing_dropdown.getChoices(),
                            smoothing_window=self.smoothing_time_eb.entry_get,
                            animal_names=self.animal_name_entry_boxes,
                            data_path=self.data_dir_select.folder_path,
                        ),
                    )

                elif data_type_choice == "H5 (SLEAP)":
                    self.data_dir_select = FolderSelect(
                        self.data_dir_frm, "H5 SLEAP DIRECTORY", lblwidth=25
                    )
                    self.instructions_lbl = Label(
                        self.data_dir_frm,
                        text="Please import videos before importing the sleap h5 tracking data \n if you are tracking more than ONE animal",
                    )
                    self.run_btn = Button(
                        self.import_frm,
                        text="IMPORT SLEAP H5",
                        fg="blue",
                        command=lambda: run_call(
                            data_type=data_type_choice,
                            interpolation=self.interpolation_dropdown.getChoices(),
                            smoothing=self.smoothing_dropdown.getChoices(),
                            smoothing_window=self.smoothing_time_eb.entry_get,
                            animal_names=self.animal_name_entry_boxes,
                            data_path=self.data_dir_select.folder_path,
                        ),
                    )

                self.data_dir_frm.grid(
                    row=self.frame_children(frame=self.choice_frm), column=0, sticky=NW
                )
                self.data_dir_select.grid(row=0, column=0, sticky=NW)
                self.instructions_lbl.grid(row=1, column=0, sticky=NW)
                self.import_frm.grid(
                    row=self.frame_children(frame=self.choice_frm) + 1,
                    column=0,
                    sticky=NW,
                )
                self.run_btn.grid(row=0, column=0, sticky=NW)
            self.choice_frm.grid(row=1, column=0, sticky=NW)

        self.import_tracking_frm = LabelFrame(
            parent_frm,
            text="IMPORT TRACKING DATA",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="black",
        )
        if not hasattr(self, "config_path"):
            self.instructions_lbl = Label(
                self.import_tracking_frm,
                text="Please CREATE PROJECT CONFIG before importing tracking data \n",
            )
            self.import_tracking_frm.grid(row=0, column=0, sticky=NW)
            self.instructions_lbl.grid(row=0, column=0, sticky=NW)
        else:
            self.config = read_config_file(config_path=self.config_path)
            self.project_animal_cnt = read_config_entry(
                config=self.config,
                section=ConfigKey.GENERAL_SETTINGS.value,
                option=ConfigKey.ANIMAL_CNT.value,
                data_type="int",
            )
            self.data_type_dropdown = DropDownMenu(
                self.import_tracking_frm,
                "DATA TYPE:",
                Options.IMPORT_TYPE_OPTIONS.value,
                labelwidth=25,
                com=import_menu,
            )
            self.data_type_dropdown.setChoices(Options.IMPORT_TYPE_OPTIONS.value[0])
            import_menu(data_type_choice=Options.IMPORT_TYPE_OPTIONS.value[0])
            self.import_tracking_frm.grid(row=idx_row, column=idx_column, sticky=NW)
            self.data_type_dropdown.grid(row=0, column=0, sticky=NW)

    def create_import_videos_menu(
        self, parent_frm: Frame, idx_row: int = 0, idx_column: int = 0
    ):

        def run_import(multiple_videos: bool):
            if multiple_videos:
                check_if_dir_exists(in_dir=self.video_directory_select.folder_path)
                copy_multiple_videos_to_project(
                    config_path=self.config_path,
                    source=self.video_directory_select.folder_path,
                    symlink=self.multiple_videos_symlink_var.get(),
                    file_type=self.video_type.getChoices(),
                )
            else:
                check_file_exist_and_readable(
                    file_path=self.video_file_select.file_path
                )
                copy_single_video_to_project(
                    simba_ini_path=self.config_path,
                    symlink=self.single_video_symlink_var.get(),
                    source_path=self.video_file_select.file_path,
                )

        import_videos_frm = LabelFrame(
            parent_frm,
            text="IMPORT VIDEOS",
            fg="black",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
        )
        if not hasattr(self, "config_path"):
            self.instructions_lbl = Label(
                import_videos_frm,
                text="Please CREATE PROJECT CONFIG before importing VIDEOS \n",
            )
            import_videos_frm.grid(row=0, column=0, sticky=NW)
            self.instructions_lbl.grid(row=0, column=0, sticky=NW)

        else:
            import_multiple_videos_frm = LabelFrame(
                import_videos_frm, text="IMPORT MULTIPLE VIDEOS"
            )
            self.video_directory_select = FolderSelect(
                import_multiple_videos_frm, "VIDEO DIRECTORY: ", lblwidth=25
            )
            self.video_type = DropDownMenu(
                import_multiple_videos_frm,
                "VIDEO FILE FORMAT: ",
                Options.VIDEO_FORMAT_OPTIONS.value,
                "25",
            )
            self.video_type.setChoices(Options.VIDEO_FORMAT_OPTIONS.value[0])
            import_multiple_btn = Button(
                import_multiple_videos_frm,
                text="Import MULTIPLE videos",
                fg="blue",
                command=lambda: run_import(multiple_videos=True),
            )
            self.multiple_videos_symlink_var = BooleanVar(value=False)
            multiple_videos_symlink_cb = Checkbutton(
                import_multiple_videos_frm,
                text="Import SYMLINKS",
                variable=self.multiple_videos_symlink_var,
            )

            import_single_frm = LabelFrame(
                import_videos_frm, text="IMPORT SINGLE VIDEO", pady=5, padx=5
            )
            self.video_file_select = FileSelect(
                import_single_frm,
                "VIDEO PATH: ",
                title="Select a video file",
                lblwidth=25,
                file_types=[("VIDEO FILE", Options.ALL_VIDEO_FORMAT_STR_OPTIONS.value)],
            )
            import_single_btn = Button(
                import_single_frm,
                text="Import SINGLE video",
                fg="blue",
                command=lambda: run_import(multiple_videos=False),
            )
            self.single_video_symlink_var = BooleanVar(value=False)
            single_video_symlink_cb = Checkbutton(
                import_single_frm,
                text="Import SYMLINK",
                variable=self.single_video_symlink_var,
            )

            import_videos_frm.grid(row=0, column=0, sticky=NW)
            import_multiple_videos_frm.grid(row=0, sticky=W)
            self.video_directory_select.grid(row=1, sticky=W)
            self.video_type.grid(row=2, sticky=W)
            multiple_videos_symlink_cb.grid(row=3, sticky=W)
            import_multiple_btn.grid(row=4, sticky=W)

            import_single_frm.grid(row=1, column=0, sticky=NW)
            self.video_file_select.grid(row=0, sticky=W)
            single_video_symlink_cb.grid(row=1, sticky=W)
            import_single_btn.grid(row=2, sticky=W)
            import_videos_frm.grid(row=idx_row, column=idx_column, sticky=NW)

    def create_multiprocess_choice(
        self,
        parent: Frame,
        cb_text: str = "Multiprocess videos (faster)",
    ):
        self.multiprocess_var = BooleanVar(value=False)
        self.multiprocess_dropdown = DropDownMenu(
            parent, "CPU cores:", list(range(2, self.cpu_cnt)), "12"
        )
        multiprocess_cb = Checkbutton(
            parent,
            text=cb_text,
            variable=self.multiprocess_var,
            command=lambda: self.enable_dropdown_from_checkbox(
                check_box_var=self.multiprocess_var,
                dropdown_menus=[self.multiprocess_dropdown],
            ),
        )
        self.multiprocess_dropdown.setChoices(2)
        self.multiprocess_dropdown.disable()
        multiprocess_cb.grid(row=self.frame_children(frame=parent), column=0, sticky=NW)
        # print(multiprocess_dropdown.popupMenugrid_info)
        self.multiprocess_dropdown.grid(
            row=multiprocess_cb.grid_info()["row"], column=1, sticky=NW
        )

    # def quit(self, e):
    #     self.main_frm.quit()
    #     self.main_frm.destroy()
    #
    # def callback(self, url):
    #     webbrowser.open_new(url)
    #
    # def move_app(self, e):
    #     print(f'+{e.x_root}+{e.y_root}')
    #     self.main_frm.geometry(f'+{e.x_root}+{e.y_root}')
    #     #print(f'+{e.x_root}x{e.y_root}')
    #     #self.main_frm.config(width=e.x_root, height=e.y_root)
    #     #self.main_frm.update()


# test = PopUpMixin(config_path='/Users/simon/Desktop/envs/troubleshooting/two_animals_16bp_032023/project_folder/project_config.ini',
#                   title='ss')
# test.create_import_pose_menu(parent_frm=test.main_frm)

# test = PopUpMixin(config_path='/Users/simon/Desktop/envs/troubleshooting/two_animals_16bp_032023/project_folder/project_config.ini',
#                   title='ss')
# test.create_import_videos_menu(parent_frm=test.main_frm)



import math

import numpy as np


def rectangle_size_calc(rectangle_dict: dict, px_mm: float) -> dict:
    """
    Compute metric height, width and area of rectangle.

    :param dict rectangle_dict: The rectangle width and height in pixels.
    :param float px_mm: Pixels per millimeter in the video.

    :example:
    >>> rectangle_size_calc(rectangle_dict={'height': 500, 'width': 500}, px_mm=10)
    >>> {'height': 500, 'width': 500, 'height_cm': 5.0, 'width_cm': 5.0, 'area_cm': 25.0}

    """

    rectangle_dict["height_cm"] = round((rectangle_dict["height"] / px_mm) / 10, 2)
    rectangle_dict["width_cm"] = round((rectangle_dict["width"] / px_mm) / 10, 2)
    rectangle_dict["area_cm"] = round(
        rectangle_dict["width_cm"] * rectangle_dict["height_cm"], 2
    )
    return rectangle_dict


def circle_size_calc(circle_dict, px_mm) -> dict:
    """
    Compute metric radius and area of circle.

    :param dict circle_dict: The circle radius in pixels
    :param float px_mm: Pixels per millimeter in the video.

    :example:
    >>> circle_size_calc(circle_dict={'radius': 100}, px_mm=5)
    >>> {'radius': 100, 'radius_cm': 2.0, 'area_cm': 12.57}
    """

    radius_cm = round((circle_dict["radius"] / px_mm) / 10, 2)
    circle_dict["radius_cm"] = radius_cm
    circle_dict["area_cm"] = round(math.pi * (radius_cm**2), 2)
    return circle_dict


def polygon_size_calc(polygon_dict, px_mm) -> dict:
    """
    Compute metric area of polygon.

    :param dict polygon_dict: The polygon vertices as np.ndarray
    :param float px_mm: Pixels per millimeter in the video.

    :example:
    >>> polygon_size_calc(polygon_dict={'vertices': np.array([[0, 2], [200, 98], [100, 876], [10, 702]])}, px_mm=5)
    >>> {'vertices': [[  0,   2], [200,  98], [100, 876], [ 10, 702]], 'area_cm': 45.29}
    """

    y_vals = polygon_dict["vertices"][:, 0]
    x_vals = polygon_dict["vertices"][:, 1]
    poly_area_px = 0.5 * np.abs(
        np.dot(x_vals, np.roll(y_vals, 1)) - np.dot(y_vals, np.roll(x_vals, 1))
    )
    polygon_dict["area_cm"] = round((poly_area_px / px_mm) / 500, 2)

    return polygon_dict


polygon_size_calc(
    polygon_dict={"vertices": np.array([[0, 2], [200, 98], [100, 876], [10, 702]])},
    px_mm=5,
)
